from itertools import combinations
from abc import ABCMeta, abstractmethod
from skmob.utils import constants
from skmob.core.trajectorydataframe import TrajDataFrame
from tqdm import tqdm
import pandas as pd
from ..utils.utils import frequency_vector, probability_vector, date_time_precision


class Attack(object):

    # """Privacy Attack
    #
    # Abstract class for a generic attack. Defines a series of functions common to all attacks.
    # Provides basic functions to compute risk for all users in a trajectory dataframe.
    # Requires the implementation of both a matching function and an assessment function, which are attack dependant.
    #
    # Parameters
    # ----------
    # knowledge_length : int
    #     the length of the background knowledge that we want to simulate. The length of the background knowledge
    #     specifies the amount of knowledge that the adversary will use for her attack. For each individual all the
    #     combinations of points of length k will be evaluated.
    # Attributes
    # ----------
    # knowledge_length : int
    #     the length of the background knowledge that we want to simulate.
    #
    # References
    # ----------
    # .. [TIST2018] Roberto Pellungrini, Luca Pappalardo, Francesca Pratesi, and Anna Monreale. 2017. A Data Mining Approach to Assess Privacy Risk in Human Mobility Data. ACM Trans. Intell. Syst. Technol. 9, 3, Article 31 (December 2017), 27 pages. DOI: https://doi.org/10.1145/3106774
    # .. [MOB2018] Roberto Pellungrini, Luca Pappalardo, Francesca Pratesi, Anna Monreale: Analyzing Privacy Risk in Human Mobility Data. STAF Workshops 2018: 114-129
    # """
    __metaclass__ = ABCMeta

    def __init__(self, knowledge_length):
        self.knowledge_length = knowledge_length

    @property
    def knowledge_length(self):
        return self._knowledge_length

    @knowledge_length.setter
    def knowledge_length(self, val):
        if val < 1:
            raise ValueError("Parameter knowledge_length should not be less than 1")
        self._knowledge_length = val

    def _all_risks(self, traj, targets=None, force_instances=False, show_progress=False):
        # """
        # Computes risk for all the users in the data. It applies the risk function to every individual in the data.
        # If it is not required to compute the risk for the entire data, the targets parameter can be used to select
        # a portion of users to perform the calculation on.
        #
        # Parameters
        # ----------
        # traj: TrajectoryDataFrame
        #     the dataframe against which to calculate risk.
        #
        # targets : TrajectoryDataFrame or list, optional
        #     the users_id target of the attack.  They must be compatible with the trajectory data. Default values is None
        #     in which case risk is computed on all users in traj. The default is `None`.
        #
        # force_instances : boolean, optional
        #     if True, returns all possible instances of background knowledge
        #     with their respective probability of reidentification. The default is `False`.
        #
        # show_progress : boolean, optional
        #     if True, shows the progress of the computation. The default is `False`.
        #
        # Returns
        # -------
        # DataFrame
        #     a DataFrame with the privacy risk for each user, in the form (user_id, risk)
        # """
        if targets is None:
            targets = traj
        else:
            if isinstance(targets, list):
                targets = traj[traj[constants.UID].isin(targets)]
            if isinstance(targets, TrajDataFrame) or isinstance(targets, pd.DataFrame):
                targets = traj[traj[constants.UID].isin(targets[constants.UID])]
        if show_progress:
            tqdm.pandas(desc="computing risk")
            risks = targets.groupby(constants.UID).progress_apply(lambda x: self._risk(x, traj, force_instances))
        else:
            risks = targets.groupby(constants.UID).apply(lambda x: self._risk(x, traj, force_instances))
        if force_instances:
            risks = risks.droplevel(1)
            risks = risks.reset_index(drop=True)
        else:
            risks = risks.reset_index(name=constants.PRIVACY_RISK)
        return risks

    def _generate_instances(self, single_traj):
        # """
        # Return a generator to all the possible background knowledge of length k for a single user_id.
        #
        # Parameters
        # ----------
        # single_traj : TrajectoryDataFrame
        #     the dataframe of the trajectory of a single individual.
        #
        # Yields
        # ------
        # generator
        #     a generator to all the possible instances of length k. Instances are tuples with the values of the actual
        #     records in the combination.
        # """
        size = len(single_traj.index)
        if self.knowledge_length > size:
            return combinations(single_traj.values, size)
        else:
            return combinations(single_traj.values, self.knowledge_length)

    def _risk(self, single_traj, traj, force_instances=False):
        """
        # Computes the risk of reidentification of an individual with respect to the entire population in the data.
        #
        # Parameters
        # ----------
        # single_traj : TrajectoryDataFrame
        #     the dataframe of the trajectory of a single individual.
        #
        # traj : TrajectoryDataFrame
        #     the dataframe with the complete data.
        #
        # force_instances : boolean, optional
        #     if True, returns all possible instances of background knowledge
        #     with their respective probability of reidentification. The default is `False`.
        #
        # Returns
        # -------
        # float
        #     the risk for the individual, expressed as a float between 0 and 1
        # """
        instances = self._generate_instances(single_traj)
        risk = 0
        if force_instances:
            inst_data = {constants.LATITUDE: list(), constants.LONGITUDE: list(),
                         constants.DATETIME: list(), constants.UID: list(),
                         constants.INSTANCE: list(), constants.INSTANCE_ELEMENT: list(),
                         constants.PROBABILITY: list()}
            inst_id = 1
            for instance in instances:
                prob = 1.0 / traj.groupby(constants.UID).apply(lambda x: self._match(x, instance)).sum()
                elem_count = 1
                for elem in instance:
                    inst_data[constants.LATITUDE].append(elem[0])
                    inst_data[constants.LONGITUDE].append(elem[1])
                    inst_data[constants.DATETIME].append(elem[2])
                    inst_data[constants.UID].append(elem[3])
                    inst_data[constants.INSTANCE].append(inst_id)
                    inst_data[constants.INSTANCE_ELEMENT].append(elem_count)
                    inst_data[constants.PROBABILITY].append(prob)
                    elem_count += 1
                inst_id += 1
            return pd.DataFrame(inst_data)
        else:
            for instance in instances:
                prob = 1.0 / traj.groupby(constants.UID).apply(lambda x: self._match(x, instance)).sum()
                if prob > risk:
                    risk = prob
                if risk == 1.0:
                    break
            return risk

    @abstractmethod
    def assess_risk(self, traj, targets=None, force_instances=False, show_progress=False):
        # """
        # Abstract function to assess privacy risk for a TrajectoryDataFrame.
        # An attack must implement an assessing strategy. This could involve some preprocessing, for example
        # transforming the original data, and calls to the risk function.
        # If it is not required to compute the risk for the entire data, the targets parameter can be used to select
        # a portion of users to perform the assessment on.
        #
        # Parameters
        # ----------
        # traj : TrajectoryDataFrame
        #     the dataframe on which to assess privacy risk.
        #
        # targets : TrajectoryDataFrame or list, optional
        #     the users_id target of the attack.  They must be compatible with the trajectory data. Default values is None
        #     in which case risk is computed on all users in traj. The defaul is `None`.
        #
        # force_instances : boolean, optional
        #     if True, returns all possible instances of background knowledge
        #     with their respective probability of reidentification. The defaul is `False`.
        #
        # show_progress : boolean, optional
        #     if True, shows the progress of the computation. The defaul is `False`.
        #
        # Returns
        # -------
        # DataFrame
        #     a DataFrame with the privacy risk for each user, in the form (user_id, risk).
        # """
        pass

    @abstractmethod
    def _match(self, single_traj, instance):
        # """
        # Matching function for the attack. It is used to decide if an instance of background knowledge matches a certain
        # trajectory. The internal logic of an attack is represented by this function, therefore, it must be implemented
        # depending in the kind of the attack.
        #
        # Parameters
        # ----------
        # single_traj : TrajectoryDataFrame
        #     the dataframe of the trajectory of a single individual.
        #
        # instance : tuple
        #     an instance of background knowledge.
        #
        # Returns
        # -------
        # int
        #     1 if the instance matches the trajectory, 0 otherwise.
        # """
        pass


class LocationAttack(Attack):
    """Location Attack

    In a location attack the adversary knows the coordinates of the locations visited by an individual and matches them
    against trajectories.

    Parameters
    ----------
    knowledge_length : int
        the length of the background knowledge that we want to simulate. The length of the background knowledge
        specifies the amount of knowledge that the adversary will use for her attack. For each individual all the
        combinations of points of length k will be evaluated.

    Attributes
    ----------
    knowledge_length : int
        the length of the background knowledge that we want to simulate.

    Examples
    --------
    >>> import skmob
    >>> from skmob.privacy import attacks
    >>> from skmob.core.trajectorydataframe import TrajDataFrame
    >>> # load data
    >>> url_priv_ex = "https://raw.githubusercontent.com/scikit-mobility/scikit-mobility/master/tutorial/data/privacy_toy.csv"
    >>> trjdat = TrajDataFrame.from_file(filename=url_priv_ex)
    >>> # create a location attack and assess risk
    >>> at = attacks.LocationAttack(knowledge_length=2)
    >>> r = at.assess_risk(trjdat)
    >>> print(r)
    	uid	risk
    0	1	0.333333
    1	2	0.500000
    2	3	0.333333
    3	4	0.333333
    4	5	0.250000
    5	6	0.250000
    6	7	0.500000

    >>> # change the length of the background knowledge and reassess risk
    >>> at.knowledge_length = 3
    >>> r = at.assess_risk(trjdat)
    >>> print(r)
       uid      risk
    0    1  0.500000
    1    2  1.000000
    2    3  0.500000
    3    4  0.333333
    4    5  0.333333
    5    6  0.250000
    6    7  1.000000

    >>> # limit privacy assessment to some target uids
    >>> r = at.assess_risk(trjdat, targets=[1,2])
    >>> print(r)
       uid  risk
    0    1   0.5
    1    2   1.0

    >>> # inspect probability of reidentification for each background knowledge instance
    >>> r = at.assess_risk(trjdat, targets=[1,2], force_instances=True)
    >>> print(r)
              lat        lng            datetime  uid  instance  instance_elem      prob
    0   43.843014  10.507994 2011-02-03 08:34:04    1         1              1  0.333333
    1   43.544270  10.326150 2011-02-03 09:34:04    1         1              2  0.333333
    2   43.708530  10.403600 2011-02-03 10:34:04    1         1              3  0.333333
    3   43.843014  10.507994 2011-02-03 08:34:04    1         2              1  0.500000
    4   43.544270  10.326150 2011-02-03 09:34:04    1         2              2  0.500000
    5   43.779250  11.246260 2011-02-04 10:34:04    1         2              3  0.500000
    6   43.843014  10.507994 2011-02-03 08:34:04    1         3              1  0.333333
    7   43.708530  10.403600 2011-02-03 10:34:04    1         3              2  0.333333
    8   43.779250  11.246260 2011-02-04 10:34:04    1         3              3  0.333333
    9   43.544270  10.326150 2011-02-03 09:34:04    1         4              1  0.333333
    10  43.708530  10.403600 2011-02-03 10:34:04    1         4              2  0.333333
    11  43.779250  11.246260 2011-02-04 10:34:04    1         4              3  0.333333
    12  43.843014  10.507994 2011-02-03 08:34:04    2         1              1  1.000000
    13  43.708530  10.403600 2011-02-03 09:34:04    2         1              2  1.000000
    14  43.843014  10.507994 2011-02-04 10:34:04    2         1              3  1.000000
    15  43.843014  10.507994 2011-02-03 08:34:04    2         2              1  0.333333
    16  43.708530  10.403600 2011-02-03 09:34:04    2         2              2  0.333333
    17  43.544270  10.326150 2011-02-04 11:34:04    2         2              3  0.333333
    18  43.843014  10.507994 2011-02-03 08:34:04    2         3              1  1.000000
    19  43.843014  10.507994 2011-02-04 10:34:04    2         3              2  1.000000
    20  43.544270  10.326150 2011-02-04 11:34:04    2         3              3  1.000000
    21  43.708530  10.403600 2011-02-03 09:34:04    2         4              1  0.333333
    22  43.843014  10.507994 2011-02-04 10:34:04    2         4              2  0.333333
    23  43.544270  10.326150 2011-02-04 11:34:04    2         4              3  0.333333


    References
    ----------
    .. [TIST2018] Roberto Pellungrini, Luca Pappalardo, Francesca Pratesi, and Anna Monreale. 2017. A Data Mining Approach to Assess Privacy Risk in Human Mobility Data. ACM Trans. Intell. Syst. Technol. 9, 3, Article 31 (December 2017), 27 pages. DOI: https://doi.org/10.1145/3106774
    .. [MOB2018] Roberto Pellungrini, Luca Pappalardo, Francesca Pratesi, Anna Monreale: Analyzing Privacy Risk in Human Mobility Data. STAF Workshops 2018: 114-129
    """

    def __init__(self, knowledge_length):
        super(LocationAttack, self).__init__(knowledge_length)

    def assess_risk(self, traj, targets=None, force_instances=False, show_progress=False):
        """
        Assess privacy risk for a TrajectoryDataFrame.
        An attack must implement an assessing strategy. This could involve some preprocessing, for example
        transforming the original data, and calls to the risk function.
        If it is not required to compute the risk for the entire data, the targets parameter can be used to select
        a portion of users to perform the assessment on.

        Parameters
        ----------
        traj : TrajectoryDataFrame
            the dataframe on which to assess privacy risk.

        targets : TrajectoryDataFrame or list, optional
            the users_id target of the attack.  They must be compatible with the trajectory data. Default values is None
            in which case risk is computed on all users in traj. The defaul is `None`.

        force_instances : boolean, optional
            if True, returns all possible instances of background knowledge
            with their respective probability of reidentification. The defaul is `False`.

        show_progress : boolean, optional
            if True, shows the progress of the computation. The defaul is `False`.

        Returns
        -------
        DataFrame
            a DataFrame with the privacy risk for each user, in the form (user_id, risk).
        """
        traj = traj.sort_values(by=[constants.UID, constants.DATETIME])
        return self._all_risks(traj, targets, force_instances, show_progress)

    def _match(self, single_traj, instance):
        """
        Matching function for the attack.
        For a location attack, only the coordinates are used in the matching.
        If a trajectory presents the same locations as the ones in the instance, a match is found.
        Multiple visits to the same location are also handled.

        Parameters
        ----------
        single_traj : TrajectoryDataFrame
            the dataframe of the trajectory of a single individual.

        instance : tuple
            an instance of background knowledge.

        Returns
        -------
        int
            1 if the instance matches the trajectory, 0 otherwise.
        """
        locs = single_traj.groupby([constants.LATITUDE, constants.LONGITUDE]).size().reset_index(name=constants.COUNT)
        inst = pd.DataFrame(data=instance, columns=single_traj.columns)
        inst = inst.astype(dtype=dict(single_traj.dtypes))
        inst = inst.groupby([constants.LATITUDE, constants.LONGITUDE]).size().reset_index(name=constants.COUNT + "inst")
        locs_inst = pd.merge(locs, inst, left_on=[constants.LATITUDE, constants.LONGITUDE],
                             right_on=[constants.LATITUDE, constants.LONGITUDE])
        if len(locs_inst.index) != len(inst.index):
            return 0
        else:
            condition = locs_inst[constants.COUNT] >= locs_inst[constants.COUNT + "inst"]
            if len(locs_inst[condition].index) != len(inst.index):
                return 0
            else:
                return 1


class LocationSequenceAttack(Attack):
    """Location Sequence Attack
    In a location sequence attack the adversary knows the coordinates of locations visited by an individual and
    the order in which they were visited and matches them against trajectories.

    Parameters
    ----------
    knowledge_length : int
        the length of the background knowledge that we want to simulate. The length of the background knowledge
        specifies the amount of knowledge that the adversary will use for her attack. For each individual all the
        combinations of points of length k will be evaluated.
    Attributes
    ----------
    knowledge_length : int
        the length of the background knowledge that we want to simulate.

    Examples
    --------
    >>> import skmob
    >>> from skmob.privacy import attacks
    >>> from skmob.core.trajectorydataframe import TrajDataFrame
    >>> # load data
    >>> url_priv_ex = "https://raw.githubusercontent.com/scikit-mobility/scikit-mobility/master/tutorial/data/privacy_toy.csv"
    >>> trjdat = TrajDataFrame.from_file(filename=url_priv_ex)
    >>> # create a location attack and assess risk
    >>> at = attacks.LocationSequenceAttack(knowledge_length=2)
    >>> r = at.assess_risk(trjdat)
    >>> print(r)
       uid      risk
    0    1  0.500000
    1    2  0.500000
    2    3  1.000000
    3    4  0.500000
    4    5  1.000000
    5    6  0.333333
    6    7  0.500000

    >>> # change the length of the background knowledge and reassess risk
    >>> at.knowledge_length = 3
    >>> r = at.assess_risk(trjdat)
    >>> print(r)
       uid      risk
    0    1  1.000000
    1    2  1.000000
    2    3  1.000000
    3    4  1.000000
    4    5  1.000000
    5    6  0.333333
    6    7  1.000000

    >>> # limit privacy assessment to some target uids
    >>> r = at.assess_risk(trjdat, targets=[1,2])
    >>> print(r)
       uid  risk
    0    1   1.0
    1    2   1.0

    >>> # inspect probability of reidentification for each background knowledge instance
    >>> r = at.assess_risk(trjdat, targets=[1,2], force_instances=True)
    >>> print(r)
              lat        lng            datetime  uid  instance  instance_elem  prob
    0   43.843014  10.507994 2011-02-03 08:34:04    1         1              1   1.0
    1   43.544270  10.326150 2011-02-03 09:34:04    1         1              2   1.0
    2   43.708530  10.403600 2011-02-03 10:34:04    1         1              3   1.0
    3   43.843014  10.507994 2011-02-03 08:34:04    1         2              1   1.0
    4   43.544270  10.326150 2011-02-03 09:34:04    1         2              2   1.0
    5   43.779250  11.246260 2011-02-04 10:34:04    1         2              3   1.0
    6   43.843014  10.507994 2011-02-03 08:34:04    1         3              1   1.0
    7   43.708530  10.403600 2011-02-03 10:34:04    1         3              2   1.0
    8   43.779250  11.246260 2011-02-04 10:34:04    1         3              3   1.0
    9   43.544270  10.326150 2011-02-03 09:34:04    1         4              1   0.5
    10  43.708530  10.403600 2011-02-03 10:34:04    1         4              2   0.5
    11  43.779250  11.246260 2011-02-04 10:34:04    1         4              3   0.5
    12  43.843014  10.507994 2011-02-03 08:34:04    2         1              1   1.0
    13  43.708530  10.403600 2011-02-03 09:34:04    2         1              2   1.0
    14  43.843014  10.507994 2011-02-04 10:34:04    2         1              3   1.0
    15  43.843014  10.507994 2011-02-03 08:34:04    2         2              1   1.0
    16  43.708530  10.403600 2011-02-03 09:34:04    2         2              2   1.0
    17  43.544270  10.326150 2011-02-04 11:34:04    2         2              3   1.0
    18  43.843014  10.507994 2011-02-03 08:34:04    2         3              1   1.0
    19  43.843014  10.507994 2011-02-04 10:34:04    2         3              2   1.0
    20  43.544270  10.326150 2011-02-04 11:34:04    2         3              3   1.0
    21  43.708530  10.403600 2011-02-03 09:34:04    2         4              1   1.0
    22  43.843014  10.507994 2011-02-04 10:34:04    2         4              2   1.0
    23  43.544270  10.326150 2011-02-04 11:34:04    2         4              3   1.0

    References
    ----------
    .. [TIST2018] Roberto Pellungrini, Luca Pappalardo, Francesca Pratesi, and Anna Monreale. 2017. A Data Mining Approach to Assess Privacy Risk in Human Mobility Data. ACM Trans. Intell. Syst. Technol. 9, 3, Article 31 (December 2017), 27 pages. DOI: https://doi.org/10.1145/3106774
    .. [MOB2018] Roberto Pellungrini, Luca Pappalardo, Francesca Pratesi, Anna Monreale: Analyzing Privacy Risk in Human Mobility Data. STAF Workshops 2018: 114-129
    """

    def __init__(self, knowledge_length):
        super(LocationSequenceAttack, self).__init__(knowledge_length)

    def assess_risk(self, traj, targets=None, force_instances=False, show_progress=False):
        """
        Assess privacy risk for a TrajectoryDataFrame.
        An attack must implement an assessing strategy. This could involve some preprocessing, for example
        transforming the original data, and calls to the risk function.
        If it is not required to compute the risk for the entire data, the targets parameter can be used to select
        a portion of users to perform the assessment on.

        Parameters
        ----------
        traj : TrajectoryDataFrame
            the dataframe on which to assess privacy risk.

        targets : TrajectoryDataFrame or list, optional
            the users_id target of the attack.  They must be compatible with the trajectory data. Default values is None
            in which case risk is computed on all users in traj. The defaul is `None`.

        force_instances : boolean, optional
            if True, returns all possible instances of background knowledge
            with their respective probability of reidentification. The defaul is `False`.

        show_progress : boolean, optional
            if True, shows the progress of the computation. The defaul is `False`.

        Returns
        -------
        DataFrame
            a DataFrame with the privacy risk for each user, in the form (user_id, risk).
        """
        traj = traj.sort_values(by=[constants.UID, constants.DATETIME])
        return self._all_risks(traj, targets, force_instances, show_progress)

    def _match(self, single_traj, instance):
        """
        Matching function for the attack.
        For a location sequence attack, both the coordinates and the order of visit are used in the matching.
        If a trajectory presents the same locations in the same order as the ones in the instance, a match is found.

        Parameters
        ----------
        single_traj : TrajectoryDataFrame
            the dataframe of the trajectory of a single individual.

        instance : tuple
            an instance of background knowledge.

        Returns
        -------
        int
            1 if the instance matches the trajectory, 0 otherwise.
        """
        inst = pd.DataFrame(data=instance, columns=single_traj.columns)
        inst_iterator = inst.iterrows()
        inst_line = next(inst_iterator)[1]
        count = 0
        for index, row in single_traj.iterrows():
            if inst_line[constants.LATITUDE] == row[constants.LATITUDE] and inst_line[constants.LONGITUDE] == row[
                constants.LONGITUDE]:
                count += 1
                try:
                    inst_line = next(inst_iterator)[1]
                except StopIteration:
                    break
        if len(inst.index) == count:
            return 1
        else:
            return 0


class LocationTimeAttack(Attack):
    """Location Time Attack

    In a location time attack the adversary knows the coordinates of locations visited by an individual and the time
    in which they were visited and matches them against trajectories. The precision at which to consider the temporal
    information can also be specified.

    Parameters
    ----------
    knowledge_length : int
        the length of the background knowledge that we want to simulate. The length of the background knowledge
        specifies the amount of knowledge that the adversary will use for her attack. For each individual all the
        combinations of points of length k will be evaluated.

    time_precision : string, optional
        the precision at which to consider the timestamps for the visits.
        The possible precisions are: Year, Month, Day, Hour, Minute, Second. The default is `Hour`

    Attributes
    ----------
    knowledge_length : int
        the length of the background knowledge that we want to simulate.

    time_precision : string
        the precision at which to consider the timestamps for the visits.

    Examples
    --------
    >>> import skmob
    >>> from skmob.privacy import attacks
    >>> from skmob.core.trajectorydataframe import TrajDataFrame
    >>> # load data
    >>> url_priv_ex = "https://raw.githubusercontent.com/scikit-mobility/scikit-mobility/master/tutorial/data/privacy_toy.csv"
    >>> trjdat = TrajDataFrame.from_file(filename=url_priv_ex)
    >>> # create a location attack and assess risk
    >>> at = attacks.LocationTimeAttack(knowledge_length=2)
    >>> r = at.assess_risk(trjdat)
    >>> print(r)
       uid  risk
    0    1   1.0
    1    2   1.0
    2    3   1.0
    3    4   1.0
    4    5   1.0
    5    6   0.5
    6    7   1.0

    >>> #change the time granularity of the attack
    >>> at.time_precision = "Month"
    >>> r = at.assess_risk(trjdat)
    >>> print(r)
       uid      risk
    0    1  0.333333
    1    2  0.500000
    2    3  0.333333
    3    4  0.333333
    4    5  0.250000
    5    6  0.250000
    6    7  0.500000

    >>> # change the length of the background knowledge and reassess risk
    >>> at.knowledge_length = 3
    >>> r = at.assess_risk(trjdat)
    >>> print(r)
       uid      risk
    0    1  0.500000
    1    2  1.000000
    2    3  0.500000
    3    4  0.333333
    4    5  0.333333
    5    6  0.250000
    6    7  1.000000

    >>> # limit privacy assessment to some target uids
    >>> r = at.assess_risk(trjdat, targets=[1,2])
    >>> print(r)
       uid      risk
    0    1  0.500000
    1    2  1.000000

    >>> # inspect probability of reidentification for each background knowledge instance
    >>> r = at.assess_risk(trjdat, targets=[1,2], force_instances=True)
    >>> print(r)
              lat        lng            datetime  uid  instance  instance_elem      prob
    0   43.843014  10.507994 2011-02-03 08:34:04    1         1              1  0.333333
    1   43.544270  10.326150 2011-02-03 09:34:04    1         1              2  0.333333
    2   43.708530  10.403600 2011-02-03 10:34:04    1         1              3  0.333333
    3   43.843014  10.507994 2011-02-03 08:34:04    1         2              1  0.500000
    4   43.544270  10.326150 2011-02-03 09:34:04    1         2              2  0.500000
    5   43.779250  11.246260 2011-02-04 10:34:04    1         2              3  0.500000
    6   43.843014  10.507994 2011-02-03 08:34:04    1         3              1  0.333333
    7   43.708530  10.403600 2011-02-03 10:34:04    1         3              2  0.333333
    8   43.779250  11.246260 2011-02-04 10:34:04    1         3              3  0.333333
    9   43.544270  10.326150 2011-02-03 09:34:04    1         4              1  0.333333
    10  43.708530  10.403600 2011-02-03 10:34:04    1         4              2  0.333333
    11  43.779250  11.246260 2011-02-04 10:34:04    1         4              3  0.333333
    12  43.843014  10.507994 2011-02-03 08:34:04    2         1              1  1.000000
    13  43.708530  10.403600 2011-02-03 09:34:04    2         1              2  1.000000
    14  43.843014  10.507994 2011-02-04 10:34:04    2         1              3  1.000000
    15  43.843014  10.507994 2011-02-03 08:34:04    2         2              1  0.333333
    16  43.708530  10.403600 2011-02-03 09:34:04    2         2              2  0.333333
    17  43.544270  10.326150 2011-02-04 11:34:04    2         2              3  0.333333
    18  43.843014  10.507994 2011-02-03 08:34:04    2         3              1  1.000000
    19  43.843014  10.507994 2011-02-04 10:34:04    2         3              2  1.000000
    20  43.544270  10.326150 2011-02-04 11:34:04    2         3              3  1.000000
    21  43.708530  10.403600 2011-02-03 09:34:04    2         4              1  0.333333
    22  43.843014  10.507994 2011-02-04 10:34:04    2         4              2  0.333333
    23  43.544270  10.326150 2011-02-04 11:34:04    2         4              3  0.333333

    References
    ----------
    .. [TIST2018] Roberto Pellungrini, Luca Pappalardo, Francesca Pratesi, and Anna Monreale. 2017. A Data Mining Approach to Assess Privacy Risk in Human Mobility Data. ACM Trans. Intell. Syst. Technol. 9, 3, Article 31 (December 2017), 27 pages. DOI: https://doi.org/10.1145/3106774
    .. [MOB2018] Roberto Pellungrini, Luca Pappalardo, Francesca Pratesi, Anna Monreale: Analyzing Privacy Risk in Human Mobility Data. STAF Workshops 2018: 114-129
    """

    def __init__(self, knowledge_length, time_precision="Hour"):
        self.time_precision = time_precision
        super(LocationTimeAttack, self).__init__(knowledge_length)

    @property
    def time_precision(self):
        return self._time_precision

    @time_precision.setter
    def time_precision(self, val):
        if val not in constants.PRECISION_LEVELS:
            raise ValueError("Possible time precisions are: Year, Month, Day, Hour, Minute, Second")
        self._time_precision = val

    def assess_risk(self, traj, targets=None, force_instances=False, show_progress=False):
        """
        Assess privacy risk for a TrajectoryDataFrame.
        An attack must implement an assessing strategy. This could involve some preprocessing, for example
        transforming the original data, and calls to the risk function.
        If it is not required to compute the risk for the entire data, the targets parameter can be used to select
        a portion of users to perform the assessment on.

        Parameters
        ----------
        traj : TrajectoryDataFrame
            the dataframe on which to assess privacy risk.

        targets : TrajectoryDataFrame or list, optional
            the users_id target of the attack.  They must be compatible with the trajectory data. Default values is None
            in which case risk is computed on all users in traj. The defaul is `None`.

        force_instances : boolean, optional
            if True, returns all possible instances of background knowledge
            with their respective probability of reidentification. The defaul is `False`.

        show_progress : boolean, optional
            if True, shows the progress of the computation. The defaul is `False`.

        Returns
        -------
        DataFrame
            a DataFrame with the privacy risk for each user, in the form (user_id, risk).
        """
        traj = traj.sort_values(by=[constants.UID, constants.DATETIME])
        traj[constants.TEMP] = traj[constants.DATETIME].apply(lambda x: date_time_precision(x, self.time_precision))
        return self._all_risks(traj, targets, force_instances, show_progress)

    def _match(self, single_traj, instance):
        """
        Matching function for the attack.
        For a location time attack, both the coordinates and the order of visit are used in the matching.
        If a trajectory presents the same locations with the same temporal information as in the instance,
        a match is found.

        Parameters
        ----------
        single_traj : TrajectoryDataFrame
            the dataframe of the trajectory of a single individual.

        instance : tuple
            an instance of background knowledge.

        Returns
        -------
        int
            1 if the instance matches the trajectory, 0 otherwise.
        """
        locs = single_traj.groupby([constants.LATITUDE, constants.LONGITUDE, constants.TEMP]).size().reset_index(name=constants.COUNT)
        inst = pd.DataFrame(data=instance, columns=single_traj.columns)
        inst = inst.groupby([constants.LATITUDE, constants.LONGITUDE,constants.TEMP]).size().reset_index(name=constants.COUNT + "inst")
        locs_inst = pd.merge(locs, inst, left_on=[constants.LATITUDE, constants.LONGITUDE, constants.TEMP],
                             right_on=[constants.LATITUDE, constants.LONGITUDE,constants.TEMP])
        if len(locs_inst.index) != len(inst.index):
            return 0
        else:
            condition = locs_inst[constants.COUNT] >= locs_inst[constants.COUNT + "inst"]
            if len(locs_inst[condition].index) != len(inst.index):
                return 0
            else:
                return 1


class UniqueLocationAttack(Attack):
    """Unique Location Attack

    In a unique location attack the adversary knows the coordinates of unique locations visited by an individual,
    and matches them against frequency vectors. A frequency vector, is an aggregation on trajectory
    data showing the unique locations visited by an individual and the frequency with which he visited those locations.

    Parameters
    ----------
    knowledge_length : int
        the length of the background knowledge that we want to simulate. The length of the background knowledge
        specifies the amount of knowledge that the adversary will use for her attack. For each individual all the
        combinations of points of length k will be evaluated.

    Attributes
    ----------
    knowledge_length : int
        the length of the background knowledge that we want to simulate.

    Examples
    --------
    >>> import skmob
    >>> from skmob.privacy import attacks
    >>> from skmob.core.trajectorydataframe import TrajDataFrame
    >>> # load data
    >>> url_priv_ex = "https://raw.githubusercontent.com/scikit-mobility/scikit-mobility/master/tutorial/data/privacy_toy.csv"
    >>> trjdat = TrajDataFrame.from_file(filename=url_priv_ex)
    >>> # create a location attack and assess risk
    >>> at = attacks.UniqueLocationAttack(knowledge_length=2)
    >>> r = at.assess_risk(trjdat)
    >>> print(r)
       uid      risk
    0    1  0.333333
    1    2  0.250000
    2    3  0.333333
    3    4  0.333333
    4    5  0.250000
    5    6  0.250000
    6    7  0.250000

    >>> # change the length of the background knowledge and reassess risk
    >>> at.knowledge_length = 3
    >>> r = at.assess_risk(trjdat)
    >>> print(r)
       uid      risk
    0    1  0.500000
    1    2  0.333333
    2    3  0.500000
    3    4  0.333333
    4    5  0.333333
    5    6  0.250000
    6    7  0.250000

    >>> # limit privacy assessment to some target uids
    >>> r = at.assess_risk(trjdat, targets=[1,2])
    >>> print(r)
       uid      risk
    0    1  0.500000
    1    2  0.333333

    >>> # inspect probability of reidentification for each background knowledge instance
    >>> r = at.assess_risk(trjdat, targets=[1,2], force_instances=True)
    >>> print(r)
        lat        lng   datetime  uid  instance  instance_elem      prob
    0   1.0  43.544270  10.326150  1.0         1              1  0.333333
    1   1.0  43.708530  10.403600  1.0         1              2  0.333333
    2   1.0  43.779250  11.246260  1.0         1              3  0.333333
    3   1.0  43.544270  10.326150  1.0         2              1  0.333333
    4   1.0  43.708530  10.403600  1.0         2              2  0.333333
    5   1.0  43.843014  10.507994  1.0         2              3  0.333333
    6   1.0  43.544270  10.326150  1.0         3              1  0.500000
    7   1.0  43.779250  11.246260  1.0         3              2  0.500000
    8   1.0  43.843014  10.507994  1.0         3              3  0.500000
    9   1.0  43.708530  10.403600  1.0         4              1  0.333333
    10  1.0  43.779250  11.246260  1.0         4              2  0.333333
    11  1.0  43.843014  10.507994  1.0         4              3  0.333333
    12  2.0  43.544270  10.326150  1.0         1              1  0.333333
    13  2.0  43.708530  10.403600  1.0         1              2  0.333333
    14  2.0  43.843014  10.507994  2.0         1              3  0.333333

    References
    ----------
    .. [TIST2018] Roberto Pellungrini, Luca Pappalardo, Francesca Pratesi, and Anna Monreale. 2017. A Data Mining Approach to Assess Privacy Risk in Human Mobility Data. ACM Trans. Intell. Syst. Technol. 9, 3, Article 31 (December 2017), 27 pages. DOI: https://doi.org/10.1145/3106774
    .. [MOB2018] Roberto Pellungrini, Luca Pappalardo, Francesca Pratesi, Anna Monreale: Analyzing Privacy Risk in Human Mobility Data. STAF Workshops 2018: 114-129
    """

    def __init__(self, knowledge_length):
        super(UniqueLocationAttack, self).__init__(knowledge_length)

    def assess_risk(self, traj, targets=None, force_instances=False, show_progress=False):
        """
        Assess privacy risk for a TrajectoryDataFrame.
        An attack must implement an assessing strategy. This could involve some preprocessing, for example
        transforming the original data, and calls to the risk function.
        If it is not required to compute the risk for the entire data, the targets parameter can be used to select
        a portion of users to perform the assessment on.

        Parameters
        ----------
        traj : TrajectoryDataFrame
            the dataframe on which to assess privacy risk.

        targets : TrajectoryDataFrame or list, optional
            the users_id target of the attack.  They must be compatible with the trajectory data. Default values is None
            in which case risk is computed on all users in traj. The defaul is `None`.

        force_instances : boolean, optional
            if True, returns all possible instances of background knowledge
            with their respective probability of reidentification. The defaul is `False`.

        show_progress : boolean, optional
            if True, shows the progress of the computation. The defaul is `False`.

        Returns
        -------
        DataFrame
            a DataFrame with the privacy risk for each user, in the form (user_id, risk).
        """
        freq = frequency_vector(traj)
        return self._all_risks(freq, targets, force_instances, show_progress)

    def _match(self, single_traj, instance):
        """
        Matching function for the attack.
        For a unique location attack, the coordinates of unique locations are used in the matching.
        If a frequency vector presents the same locations as in the instance, a match is found.

        Parameters
        ----------
        single_traj : TrajectoryDataFrame
            the dataframe of the trajectory of a single individual.

        instance : tuple
            an instance of background knowledge.

        Returns
        -------
        int
            1 if the instance matches the trajectory, 0 otherwise.
        """
        inst = pd.DataFrame(data=instance, columns=single_traj.columns)
        locs_inst = pd.merge(single_traj, inst, left_on=[constants.LATITUDE, constants.LONGITUDE],
                             right_on=[constants.LATITUDE, constants.LONGITUDE])
        if len(locs_inst.index) == len(inst.index):
            return 1
        else:
            return 0


class LocationFrequencyAttack(Attack):
    """Location Frequency Attack

    In a location frequency attack the adversary knows the coordinates of the unique locations visited by an individual
    and the frequency with which he visited them, and matches them against frequency vectors. A frequency vector,
    is an aggregation on trajectory data showing the unique locations visited by an individual and the frequency
    with which he visited those locations. It is possible to specify a tolerance level for the matching of the frequency.

    Parameters
    ----------
    knowledge_length : int
        the length of the background knowledge that we want to simulate. The length of the background knowledge
        specifies the amount of knowledge that the adversary will use for her attack. For each individual all the
        combinations of points of length k will be evaluated.

    tolerance : float, optional
        the tolarance with which to match the frequency. It can assume values between 0 and 1. The defaul is `0`.

    Attributes
    ----------
    knowledge_length : int
        the length of the background knowledge that we want to simulate.

    tolerance : float
        the tolarance with which to match the frequency.

    Examples
    --------
    >>> import skmob
    >>> from skmob.privacy import attacks
    >>> from skmob.core.trajectorydataframe import TrajDataFrame
    >>> # load data
    >>> url_priv_ex = "https://raw.githubusercontent.com/scikit-mobility/scikit-mobility/master/tutorial/data/privacy_toy.csv"
    >>> trjdat = TrajDataFrame.from_file(filename=url_priv_ex)
    >>> # create a location attack and assess risk
    >>> at = attacks.LocationFrequencyAttack(knowledge_length=2)
    >>> r = at.assess_risk(trjdat)
    >>> print(r)
       uid      risk
    0    1  0.333333
    1    2  1.000000
    2    3  0.333333
    3    4  0.333333
    4    5  0.333333
    5    6  0.333333
    6    7  1.000000

    >>> # change the tolerance with witch the frequency is matched
    >>> at.tolerance = 0.5
    >>> r = at.assess_risk(trjdat)
    >>> print(r)
       uid      risk
    0    1  0.333333
    1    2  1.000000
    2    3  0.333333
    3    4  0.333333
    4    5  0.250000
    5    6  0.250000
    6    7  1.000000

    >>> # change the length of the background knowledge and reassess risk
    >>> at.knowledge_length = 3
    >>> r = at.assess_risk(trjdat)
    >>> print(r)
       uid      risk
    0    1  0.500000
    1    2  1.000000
    2    3  0.500000
    3    4  0.333333
    4    5  0.333333
    5    6  0.250000
    6    7  1.000000

    >>> # limit privacy assessment to some target uids
    >>> r = at.assess_risk(trjdat, targets=[1,2])
    >>> print(r)
       uid  risk
    0    1   0.5
    1    2   1.0

    >>> # inspect probability of reidentification for each background knowledge instance
    >>> r = at.assess_risk(trjdat, targets=[1,2], force_instances=True)
    >>> print(r)
        lat        lng   datetime  uid  instance  instance_elem      prob
    0   1.0  43.544270  10.326150  1.0         1              1  0.333333
    1   1.0  43.708530  10.403600  1.0         1              2  0.333333
    2   1.0  43.779250  11.246260  1.0         1              3  0.333333
    3   1.0  43.544270  10.326150  1.0         2              1  0.333333
    4   1.0  43.708530  10.403600  1.0         2              2  0.333333
    5   1.0  43.843014  10.507994  1.0         2              3  0.333333
    6   1.0  43.544270  10.326150  1.0         3              1  0.500000
    7   1.0  43.779250  11.246260  1.0         3              2  0.500000
    8   1.0  43.843014  10.507994  1.0         3              3  0.500000
    9   1.0  43.708530  10.403600  1.0         4              1  0.333333
    10  1.0  43.779250  11.246260  1.0         4              2  0.333333
    11  1.0  43.843014  10.507994  1.0         4              3  0.333333
    12  2.0  43.544270  10.326150  1.0         1              1  1.000000
    13  2.0  43.708530  10.403600  1.0         1              2  1.000000
    14  2.0  43.843014  10.507994  2.0         1              3  1.000000

    References
    ----------
    .. [TIST2018] Roberto Pellungrini, Luca Pappalardo, Francesca Pratesi, and Anna Monreale. 2017. A Data Mining Approach to Assess Privacy Risk in Human Mobility Data. ACM Trans. Intell. Syst. Technol. 9, 3, Article 31 (December 2017), 27 pages. DOI: https://doi.org/10.1145/3106774
    .. [MOB2018] Roberto Pellungrini, Luca Pappalardo, Francesca Pratesi, Anna Monreale: Analyzing Privacy Risk in Human Mobility Data. STAF Workshops 2018: 114-129
    """

    def __init__(self, knowledge_length, tolerance=0.0):
        self.tolerance = tolerance
        super(LocationFrequencyAttack, self).__init__(knowledge_length)

    @property
    def tolerance(self):
        return self._tolerance

    @tolerance.setter
    def tolerance(self, val):
        if val > 1.0 or val < 0.0:
            raise ValueError("Tolerance should be in the interval [0.0,1.0]")
        self._tolerance = val

    def assess_risk(self, traj, targets=None, force_instances=False, show_progress=False):
        """
        Assess privacy risk for a TrajectoryDataFrame.
        An attack must implement an assessing strategy. This could involve some preprocessing, for example
        transforming the original data, and calls to the risk function.
        If it is not required to compute the risk for the entire data, the targets parameter can be used to select
        a portion of users to perform the assessment on.

        Parameters
        ----------
        traj : TrajectoryDataFrame
            the dataframe on which to assess privacy risk.

        targets : TrajectoryDataFrame or list, optional
            the users_id target of the attack.  They must be compatible with the trajectory data. Default values is None
            in which case risk is computed on all users in traj. The defaul is `None`.

        force_instances : boolean, optional
            if True, returns all possible instances of background knowledge
            with their respective probability of reidentification. The defaul is `False`.

        show_progress : boolean, optional
            if True, shows the progress of the computation. The defaul is `False`.

        Returns
        -------
        DataFrame
            a DataFrame with the privacy risk for each user, in the form (user_id, risk).
        """
        freq = frequency_vector(traj)
        return self._all_risks(freq, targets, force_instances, show_progress)

    def _match(self, single_traj, instance):
        """
        Matching function for the attack.
        For a frequency location attack, the coordinates of unique locations and their frequency of visit are used
        in the matching. If a frequency vector presents the same locations with the same frequency as in the instance,
        a match is found. The tolerance level specified at construction is used to construct and interval of frequency
        and allow for less precise matching.

        Parameters
        ----------
        single_traj : TrajectoryDataFrame
            the dataframe of the trajectory of a single individual.

        instance : tuple
            an instance of background knowledge.

        Returns
        -------
        int
            1 if the instance matches the trajectory, 0 otherwise.
        """
        inst = pd.DataFrame(data=instance, columns=single_traj.columns)
        inst.rename(columns={constants.FREQUENCY: constants.FREQUENCY + "inst"}, inplace=True)
        locs_inst = pd.merge(single_traj, inst, left_on=[constants.LATITUDE, constants.LONGITUDE],
                             right_on=[constants.LATITUDE, constants.LONGITUDE])
        if len(locs_inst.index) != len(inst.index):
            return 0
        else:
            condition1 = locs_inst[constants.FREQUENCY + "inst"] >= locs_inst[constants.FREQUENCY] - (
                    locs_inst[constants.FREQUENCY] * self.tolerance)
            condition2 = locs_inst[constants.FREQUENCY + "inst"] <= locs_inst[constants.FREQUENCY] + (
                    locs_inst[constants.FREQUENCY] * self.tolerance)
            if len(locs_inst[condition1 & condition2].index) != len(inst.index):
                return 0
            else:
                return 1


class LocationProbabilityAttack(Attack):
    """Location Probability Attack

    In a location probability attack the adversary knows the coordinates of
    the unique locations visited by an individual and the probability with which he visited them,
    and matches them against probability vectors.
    A probability vector, is an aggregation on trajectory data showing the unique locations visited by an individual
    and the probability with which he visited those locations.
    It is possible to specify a tolerance level for the matching of the probability.

    Parameters
    ----------
    knowledge_length : int
        the length of the background knowledge that we want to simulate. The length of the background knowledge
        specifies the amount of knowledge that the adversary will use for her attack. For each individual all the
        combinations of points of length k will be evaluated.

    tolerance : float, optional
        the tolarance with which to match the probability. It can assume values between 0 and 1. The defaul is `0`.

    Attributes
    ----------
    knowledge_length : int
        the length of the background knowledge that we want to simulate.

    tolerance : float
        the tolarance with which to match the probability.

    Examples
    --------
    >>> import skmob
    >>> from skmob.privacy import attacks
    >>> from skmob.core.trajectorydataframe import TrajDataFrame
    >>> # load data
    >>> url_priv_ex = "https://raw.githubusercontent.com/scikit-mobility/scikit-mobility/master/tutorial/data/privacy_toy.csv"
    >>> trjdat = TrajDataFrame.from_file(filename=url_priv_ex)
    >>> # create a location attack and assess risk
    >>> at = attacks.LocationProbabilityAttack(knowledge_length=2)
    >>> r = at.assess_risk(trjdat)
    >>> print(r)
       uid  risk
    0    1   0.5
    1    2   1.0
    2    3   0.5
    3    4   1.0
    4    5   1.0
    5    6   1.0
    6    7   1.0

    >>> # change the tolerance with witch the frequency is matched
    >>> at.tolerance = 0.5
    >>> r = at.assess_risk(trjdat)
    >>> print(r)
       uid      risk
    0    1  0.333333
    1    2  0.500000
    2    3  0.333333
    3    4  0.333333
    4    5  0.250000
    5    6  1.000000
    6    7  1.000000

    >>> # change the length of the background knowledge and reassess risk
    >>> at.knowledge_length = 3
    >>> r = at.assess_risk(trjdat)
    >>> print(r)
       uid      risk
    0    1  0.500000
    1    2  1.000000
    2    3  0.500000
    3    4  0.333333
    4    5  0.333333
    5    6  1.000000
    6    7  1.000000

    >>> # limit privacy assessment to some target uids
    >>> r = at.assess_risk(trjdat, targets=[1,2])
    >>> print(r)
       uid  risk
    0    1   0.5
    1    2   1.0
    >>> # inspect probability of reidentification for each background knowledge instance
    >>> r = at.assess_risk(trjdat, targets=[1,2], force_instances=True)
    >>> print(r)
        lat        lng   datetime   uid  instance  instance_elem      prob
    0   1.0  43.544270  10.326150  0.25         1              1  0.333333
    1   1.0  43.708530  10.403600  0.25         1              2  0.333333
    2   1.0  43.779250  11.246260  0.25         1              3  0.333333
    3   1.0  43.544270  10.326150  0.25         2              1  0.333333
    4   1.0  43.708530  10.403600  0.25         2              2  0.333333
    5   1.0  43.843014  10.507994  0.25         2              3  0.333333
    6   1.0  43.544270  10.326150  0.25         3              1  0.500000
    7   1.0  43.779250  11.246260  0.25         3              2  0.500000
    8   1.0  43.843014  10.507994  0.25         3              3  0.500000
    9   1.0  43.708530  10.403600  0.25         4              1  0.333333
    10  1.0  43.779250  11.246260  0.25         4              2  0.333333
    11  1.0  43.843014  10.507994  0.25         4              3  0.333333
    12  2.0  43.544270  10.326150  0.25         1              1  1.000000
    13  2.0  43.708530  10.403600  0.25         1              2  1.000000
    14  2.0  43.843014  10.507994  0.50         1              3  1.000000

    References
    ----------
    .. [TIST2018] Roberto Pellungrini, Luca Pappalardo, Francesca Pratesi, and Anna Monreale. 2017. A Data Mining Approach to Assess Privacy Risk in Human Mobility Data. ACM Trans. Intell. Syst. Technol. 9, 3, Article 31 (December 2017), 27 pages. DOI: https://doi.org/10.1145/3106774
    .. [MOB2018] Roberto Pellungrini, Luca Pappalardo, Francesca Pratesi, Anna Monreale: Analyzing Privacy Risk in Human Mobility Data. STAF Workshops 2018: 114-129
    """

    def __init__(self, knowledge_length, tolerance=0.0):
        self.tolerance = tolerance
        super(LocationProbabilityAttack, self).__init__(knowledge_length)

    @property
    def tolerance(self):
        return self._tolerance

    @tolerance.setter
    def tolerance(self, val):
        if val > 1.0 or val < 0.0:
            raise ValueError("Tolerance should be in the interval [0.0,1.0]")
        self._tolerance = val

    def assess_risk(self, traj, targets=None, force_instances=False, show_progress=False):
        """
        Assess privacy risk for a TrajectoryDataFrame.
        An attack must implement an assessing strategy. This could involve some preprocessing, for example
        transforming the original data, and calls to the risk function.
        If it is not required to compute the risk for the entire data, the targets parameter can be used to select
        a portion of users to perform the assessment on.

        Parameters
        ----------
        traj : TrajectoryDataFrame
            the dataframe on which to assess privacy risk.

        targets : TrajectoryDataFrame or list, optional
            the users_id target of the attack.  They must be compatible with the trajectory data. Default values is None
            in which case risk is computed on all users in traj. The defaul is `None`.

        force_instances : boolean, optional
            if True, returns all possible instances of background knowledge
            with their respective probability of reidentification. The defaul is `False`.

        show_progress : boolean, optional
            if True, shows the progress of the computation. The defaul is `False`.

        Returns
        -------
        DataFrame
            a DataFrame with the privacy risk for each user, in the form (user_id, risk).
        """
        prob = probability_vector(traj)
        return self._all_risks(prob, targets, force_instances, show_progress)

    def _match(self, single_traj, instance):
        """
        Matching function for the attack.
        For a probability location attack, the coordinates of unique locations and their probability of visit are used
        in the matching.
        If a probability vector presents the same locations with the same probability as in the instance,
        a match is found.
        The tolerance level specified at construction is used to build and interval of probability and allow
        for less precise matching.

        Parameters
        ----------
        single_traj : TrajectoryDataFrame
            the dataframe of the trajectory of a single individual.

        instance : tuple
            an instance of background knowledge.

        Returns
        -------
        int
            1 if the instance matches the trajectory, 0 otherwise.
        """
        inst = pd.DataFrame(data=instance, columns=single_traj.columns)
        inst.rename(columns={constants.PROBABILITY: constants.PROBABILITY + "inst"}, inplace=True)
        locs_inst = pd.merge(single_traj, inst, left_on=[constants.LATITUDE, constants.LONGITUDE],
                             right_on=[constants.LATITUDE, constants.LONGITUDE])
        if len(locs_inst.index) != len(inst.index):
            return 0
        else:
            condition1 = locs_inst[constants.PROBABILITY + "inst"] >= locs_inst[constants.PROBABILITY] - (
                    locs_inst[constants.PROBABILITY] * self.tolerance)
            condition2 = locs_inst[constants.PROBABILITY + "inst"] <= locs_inst[constants.PROBABILITY] + (
                    locs_inst[constants.PROBABILITY] * self.tolerance)
            if len(locs_inst[condition1 & condition2].index) != len(inst.index):
                return 0
            else:
                return 1


class LocationProportionAttack(Attack):
    """Location Proportion Attack

    In a location proportion attack the adversary knows the coordinates of the unique locations visited
    by an individual and the relative proportions between their frequencies of visit,
    and matches them against frequency vectors.
    A frequency vector is an aggregation on trajectory data showing the unique locations visited by an individual
    and the frequency with which he visited those locations.
    It is possible to specify a tolerance level for the matching of the proportion.

    Parameters
    ----------
    knowledge_length : int
        the length of the background knowledge that we want to simulate. The length of the background knowledge
        specifies the amount of knowledge that the adversary will use for her attack. For each individual all the
        combinations of points of length k will be evaluated.

    tolerance : float, optional
        the tolarance with which to match the frequency. It can assume values between 0 and 1. The defaul is `0`.

    Attributes
    ----------
    knowledge_length : int
        the length of the background knowledge that we want to simulate.

    tolerance : float
        the tolarance with which to match the frequency.

    Examples
    --------
    >>> import skmob
    >>> from skmob.privacy import attacks
    >>> from skmob.core.trajectorydataframe import TrajDataFrame
    >>> # load data
    >>> url_priv_ex = "https://raw.githubusercontent.com/scikit-mobility/scikit-mobility/master/tutorial/data/privacy_toy.csv"
    >>> trjdat = TrajDataFrame.from_file(filename=url_priv_ex)
    >>> # create a location attack and assess risk
    >>> at = attacks.LocationProportionAttack(knowledge_length=2)
    >>> r = at.assess_risk(trjdat)
    >>> print(r)
       uid      risk
    0    1  0.333333
    1    2  1.000000
    2    3  0.333333
    3    4  0.333333
    4    5  0.333333
    5    6  0.333333
    6    7  1.000000

    >>> # change the tolerance with witch the frequency is matched
    >>> at.tolerance = 0.5
    >>> r = at.assess_risk(trjdat)
    >>> print(r)
       uid      risk
    0    1  0.333333
    1    2  0.250000
    2    3  0.333333
    3    4  0.333333
    4    5  0.333333
    5    6  0.333333
    6    7  0.250000

    >>> # change the length of the background knowledge and reassess risk
    >>> at.knowledge_length = 3
    >>> r = at.assess_risk(trjdat)
    >>> print(r)
       uid      risk
    0    1  0.500000
    1    2  0.333333
    2    3  0.500000
    3    4  0.333333
    4    5  0.333333
    5    6  0.333333
    6    7  0.250000

    >>> # limit privacy assessment to some target uids
    >>> r = at.assess_risk(trjdat, targets=[1,2])
    >>> print(r)
       uid      risk
    0    1  0.500000
    1    2  0.333333

    >>> # inspect probability of reidentification for each background knowledge instance
    >>> r = at.assess_risk(trjdat, targets=[1,2], force_instances=True)
    >>> print(r)
        lat        lng   datetime  uid  instance  instance_elem      prob
    0   1.0  43.544270  10.326150  1.0         1              1  0.333333
    1   1.0  43.708530  10.403600  1.0         1              2  0.333333
    2   1.0  43.779250  11.246260  1.0         1              3  0.333333
    3   1.0  43.544270  10.326150  1.0         2              1  0.500000
    4   1.0  43.708530  10.403600  1.0         2              2  0.500000
    5   1.0  43.843014  10.507994  1.0         2              3  0.500000
    6   1.0  43.544270  10.326150  1.0         3              1  0.500000
    7   1.0  43.779250  11.246260  1.0         3              2  0.500000
    8   1.0  43.843014  10.507994  1.0         3              3  0.500000
    9   1.0  43.708530  10.403600  1.0         4              1  0.333333
    10  1.0  43.779250  11.246260  1.0         4              2  0.333333
    11  1.0  43.843014  10.507994  1.0         4              3  0.333333
    12  2.0  43.544270  10.326150  1.0         1              1  0.333333
    13  2.0  43.708530  10.403600  1.0         1              2  0.333333
    14  2.0  43.843014  10.507994  2.0         1              3  0.333333

    References
    ----------
    .. [TIST2018] Roberto Pellungrini, Luca Pappalardo, Francesca Pratesi, and Anna Monreale. 2017. A Data Mining Approach to Assess Privacy Risk in Human Mobility Data. ACM Trans. Intell. Syst. Technol. 9, 3, Article 31 (December 2017), 27 pages. DOI: https://doi.org/10.1145/3106774
    .. [MOB2018] Roberto Pellungrini, Luca Pappalardo, Francesca Pratesi, Anna Monreale: Analyzing Privacy Risk in Human Mobility Data. STAF Workshops 2018: 114-129
    """

    def __init__(self, knowledge_length, tolerance=0.0):
        self.tolerance = tolerance
        super(LocationProportionAttack, self).__init__(knowledge_length)

    @property
    def tolerance(self):
        return self._tolerance

    @tolerance.setter
    def tolerance(self, val):
        if val > 1.0 or val < 0.0:
            raise ValueError("Tolerance should be in the interval [0.0,1.0]")
        self._tolerance = val

    def assess_risk(self, traj, targets=None, force_instances=False, show_progress=False):
        """
        Assess privacy risk for a TrajectoryDataFrame.
        An attack must implement an assessing strategy. This could involve some preprocessing, for example
        transforming the original data, and calls to the risk function.
        If it is not required to compute the risk for the entire data, the targets parameter can be used to select
        a portion of users to perform the assessment on.

        Parameters
        ----------
        traj : TrajectoryDataFrame
            the dataframe on which to assess privacy risk.

        targets : TrajectoryDataFrame or list, optional
            the users_id target of the attack.  They must be compatible with the trajectory data. Default values is None
            in which case risk is computed on all users in traj. The defaul is `None`.

        force_instances : boolean, optional
            if True, returns all possible instances of background knowledge
            with their respective probability of reidentification. The defaul is `False`.

        show_progress : boolean, optional
            if True, shows the progress of the computation. The defaul is `False`.

        Returns
        -------
        DataFrame
            a DataFrame with the privacy risk for each user, in the form (user_id, risk).
        """
        freq = frequency_vector(traj)
        return self._all_risks(freq, targets, force_instances, show_progress)

    def _match(self, single_traj, instance):
        """
        Matching function for the attack. For a proportion location attack,
        the coordinates of unique locations and their relative proportion of frequency of visit
        are used in the matching.
        The proportion of visit are calculated with respect to the most frequent location found in the instance.
        If a frequency vector presents the same locations with the same proportions of frequency of
        visit as in the instance, a match is found.
        The tolerance level specified at construction is used to build an interval of proportion
        and allow for less precise matching.

        Parameters
        ----------
        single_traj : TrajectoryDataFrame
            the dataframe of the trajectory of a single individual.

        instance : tuple
            an instance of background knowledge.

        Returns
        -------
        int
            1 if the instance matches the trajectory, 0 otherwise.
        """
        inst = pd.DataFrame(data=instance, columns=single_traj.columns)
        inst.rename(columns={constants.FREQUENCY: constants.FREQUENCY + "inst"}, inplace=True)
        locs_inst = pd.merge(single_traj, inst, left_on=[constants.LATITUDE, constants.LONGITUDE],
                             right_on=[constants.LATITUDE, constants.LONGITUDE])
        if len(locs_inst.index) != len(inst.index):
            return 0
        else:
            locs_inst[constants.PROPORTION + "inst"] = locs_inst[constants.FREQUENCY + "inst"] / locs_inst[
                constants.FREQUENCY + "inst"].max()
            locs_inst[constants.PROPORTION] = locs_inst[constants.FREQUENCY] / locs_inst[constants.FREQUENCY].max()
            condition1 = locs_inst[constants.PROPORTION + "inst"] >= locs_inst[constants.PROPORTION] - (
                    locs_inst[constants.PROPORTION] * self.tolerance)
            condition2 = locs_inst[constants.PROPORTION + "inst"] <= locs_inst[constants.PROPORTION] + (
                    locs_inst[constants.PROPORTION] * self.tolerance)
            if len(locs_inst[condition1 & condition2].index) != len(inst.index):
                return 0
            else:
                return 1


class HomeWorkAttack(Attack):
    """Home And Work Attack

    In a home and work attack the adversary knows the coordinates of
    the two locations most frequently visited by an individual, and matches them against frequency vectors.
    A frequency vector is an aggregation on trajectory data showing the unique
    locations visited by an individual and the frequency with which he visited those locations.
    This attack does not require the generation of combinations to build the possible instances of background knowledge.

    Parameters
    ----------
    knowledge_length : int
        the length of the background knowledge that we want to simulate. The length of the background knowledge
        specifies the amount of knowledge that the adversary will use for her attack. For each individual all the
        combinations of points of length k will be evaluated.
    Attributes
    ----------
    knowledge_length : int
        the length of the background knowledge that we want to simulate.

    Examples
    --------
    >>> import skmob
    >>> from skmob.privacy import attacks
    >>> from skmob.core.trajectorydataframe import TrajDataFrame
    >>> # load data
    >>> url_priv_ex = "https://raw.githubusercontent.com/scikit-mobility/scikit-mobility/master/tutorial/data/privacy_toy.csv"
    >>> trjdat = TrajDataFrame.from_file(filename=url_priv_ex)
    >>> # create a location attack and assess risk
    >>> at = attacks.HomeWorkAttack()
    >>> r = at.assess_risk(trjdat)
    >>> print(r)
       uid  risk
    0    1  0.25
    1    2  0.25
    2    3  0.25
    3    4  0.25
    4    5  1.00
    5    6  1.00
    6    7  1.00

    >>> # limit privacy assessment to some target uids
    >>> r = at.assess_risk(trjdat, targets=[1,2])
    >>> print(r)
       uid  risk
    0    1  0.25
    1    2  0.25

    >>> # inspect probability of reidentification for each background knowledge instance
    >>> r = at.assess_risk(trjdat, targets=[1,2], force_instances=True)
    >>> print(r)
       lat       lng  datetime  uid  instance  instance_elem  prob
    0  1.0  43.54427  10.32615  1.0         1              1  0.25
    1  1.0  43.70853  10.40360  1.0         1              2  0.25
    2  2.0  43.54427  10.32615  1.0         1              1  0.25
    3  2.0  43.70853  10.40360  1.0         1              2  0.25

    References
    ----------
    .. [TIST2018] Roberto Pellungrini, Luca Pappalardo, Francesca Pratesi, and Anna Monreale. 2017. A Data Mining Approach to Assess Privacy Risk in Human Mobility Data. ACM Trans. Intell. Syst. Technol. 9, 3, Article 31 (December 2017), 27 pages. DOI: https://doi.org/10.1145/3106774
    .. [MOB2018] Roberto Pellungrini, Luca Pappalardo, Francesca Pratesi, Anna Monreale: Analyzing Privacy Risk in Human Mobility Data. STAF Workshops 2018: 114-129
    """

    def __init__(self, knowledge_length=1):
        super(HomeWorkAttack, self).__init__(knowledge_length)

    def _generate_instances(self, single_traj):
        """
        Returns the two most frequently visited locations by an individual.
        This is an ovverride of the _generate_instances method of the Attack absttract class.

        Parameters
        ----------
        single_traj : TrajectoryDataFrame
            the dataframe of the trajectory of a single individual.

        Returns
        -------
        list
            a list with the records with the two most frequently visited locations.
        """
        return [single_traj[:2].values]

    def assess_risk(self, traj, targets=None, force_instances=False, show_progress=False):
        """
        Assess privacy risk for a TrajectoryDataFrame.
        An attack must implement an assessing strategy. This could involve some preprocessing, for example
        transforming the original data, and calls to the risk function.
        If it is not required to compute the risk for the entire data, the targets parameter can be used to select
        a portion of users to perform the assessment on.

        Parameters
        ----------
        traj : TrajectoryDataFrame
            the dataframe on which to assess privacy risk.

        targets : TrajectoryDataFrame or list, optional
            the users_id target of the attack.  They must be compatible with the trajectory data. Default values is None
            in which case risk is computed on all users in traj. The defaul is `None`.

        force_instances : boolean, optional
            if True, returns all possible instances of background knowledge
            with their respective probability of reidentification. The defaul is `False`.

        show_progress : boolean, optional
            if True, shows the progress of the computation. The defaul is `False`.

        Returns
        -------
        DataFrame
            a DataFrame with the privacy risk for each user, in the form (user_id, risk).
        """
        freq = frequency_vector(traj)
        return self._all_risks(freq, targets, force_instances, show_progress)

    def _match(self, single_traj, instance):
        """
        Matching function for the attack.
        For a home and work attack, the coordinates of the two locations are used in the matching.
        If a frequency vector presents the same locations as in the instance, a match is found.

        Parameters
        ----------
        single_traj : TrajectoryDataFrame
            the dataframe of the trajectory of a single individual.

        instance : tuple
            an instance of background knowledge.

        Returns
        -------
        int
            1 if the instance matches the trajectory, 0 otherwise.
        """
        inst = pd.DataFrame(data=instance, columns=single_traj.columns)
        locs_inst = pd.merge(single_traj[:2], inst, left_on=[constants.LATITUDE, constants.LONGITUDE],
                             right_on=[constants.LATITUDE, constants.LONGITUDE])
        if len(locs_inst.index) == len(inst.index):
            return 1
        else:
            return 0



# Adapted from code & formulas by David Z. Creemer and others
# http://www.zachary.com/blog/2005/01/12/python_zipcode_geo-programming
# http://williams.best.vwh.net/avform.htm
#

from math import sin,cos,atan,acos,asin,atan2,sqrt,pi, modf
import csv
from skmob.utils import constants

# At the equator / on another great circle???
nauticalMilePerLat = 60.00721
nauticalMilePerLongitude = 60.10793

rad = pi / 180.0

milesPerNauticalMile = 1.15078
kmsPerNauticalMile = 1.85200

degreeInMiles = milesPerNauticalMile * 60
degreeInKms = kmsPerNauticalMile * 60

# earth's mean radius = 6,371km
earthradius = 6371.0


def getDistance(loc1, loc2):
    "aliased default algorithm; args are (lat_decimal,lon_decimal) tuples"
    return getDistanceByHaversine(loc1, loc2)


def getDistanceByHaversine(loc1, loc2):
    "Haversine formula - give coordinates as (lat_decimal,lon_decimal) tuples"

    lat1, lon1 = loc1
    lat2, lon2 = loc2
    
    # convert to radians
    lon1 = lon1 * pi / 180.0
    lon2 = lon2 * pi / 180.0
    lat1 = lat1 * pi / 180.0
    lat2 = lat2 * pi / 180.0

    # haversine formula
    dlon = lon2 - lon1
    dlat = lat2 - lat1
    a = (sin(dlat/2))**2 + cos(lat1) * cos(lat2) * (sin(dlon/2.0))**2
    c = 2.0 * atan2(sqrt(a), sqrt(1.0-a))
    km = earthradius * c
    return km


def DecimalToDMS(decimalvalue):
    "convert a decimal value to degree,minute,second tuple"
    d = modf(decimalvalue)[0]
    m=0
    s=0
    return (d,m,s)


def DMSToDecimal(degrees,minutes,seconds):
    "Convert a value from decimal (float) to degree,minute,second tuple"
    d = abs(degrees) + (minutes/60.0) + (seconds/3600.0)
    if degrees < 0:
        return -d
    else:
        return d


def getCoordinatesForDistance(originlat, originlon, distance, units="km"):
    """return longitude & latitude values that, when added to & subtraced from
    origin longitude & latitude, form a cross / 'plus sign' whose ends are
    a given distance from the origin"""

    degreelength = 0

    if units == "km":
        degreelength = degreeInKms
    elif units == "miles":
        degreelength = degreeInMiles
    else:
        raise Exception("Units must be either 'km' or 'miles'!")

    lat = distance / degreelength
    lon = distance / (cos(originlat * rad) * degreelength)

    return (lat, lon)


def isWithinDistance(origin, loc, distance):
    "boolean for checking whether a location is within a distance"
    if getDistanceByHaversine(origin, loc) <= distance:
        return True
    else:
        return False

def load_spatial_tessellation(filename='location2info_trentino', delimiter=','):
    """
    Load into a dictionary the locations and corresponding information (latitude, longitude, relevance)

    Parameters
    ----------
    filename: str
        the filename where the location info is stored

    Returns
    -------
    dict
        the dictionary of locations
    """
    spatial_tessellation = {}
    f = csv.reader(open(filename), delimiter=delimiter)
    f.__next__()  # delete header
    i = 0
    for line in f:  # tqdm print a progress bar
        relevance = int(line[2])
        spatial_tessellation[i] = {constants.LATITUDE: float(line[0]),
                                       constants.LONGITUDE: float(line[1]),
                                       'relevance': relevance}
        i += 1
    return spatial_tessellation



from ..utils import utils, constants
from ..core.trajectorydataframe import *
from sklearn.cluster import DBSCAN
import numpy as np
import inspect

kms_per_radian = 6371.0088   # Caution: this is only true at the Equator!
                             # This may cause problems at high latitudes.


def cluster(tdf, cluster_radius_km=0.1, min_samples=1):
    """Clustering of locations.
    
    Cluster the stops of each individual in a TrajDataFrame. The stops correspond to visits to the same location at different times, based on spatial proximity [RT2004]_. The clustering algorithm used is DBSCAN (by sklearn [DBSCAN]_).
    
    Parameters
    ----------
    tdf : TrajDataFrame
        the input TrajDataFrame that should contain the stops, i.e., the output of a `preprocessing.detection` function.

    cluster_radius_km : float, optional
        the parameter `eps` of the function sklearn.cluster.DBSCAN, in kilometers. The default is `0.1`.
        
    min_samples : int, optional
        the parameter `min_samples` of the function sklearn.cluster.DBSCAN indicating the minimum number of stops to form a cluster. The default is `1`.
    
    Returns
    -------
    TrajDataFrame
        a TrajDataFrame with the additional column 'cluster' containing the cluster labels. The stops that belong to the same cluster have the same label. The labels are integers corresponding to the ranks of clusters according to the frequency of visitation (the most visited cluster has label 0, the second most visited has label 1, etc.).
    
    Examples
    --------
    >>> import skmob
    >>> import pandas as pd
    >>> from skmob.preprocessing import detection, clustering
    >>> # read the trajectory data (GeoLife)
    >>> url = 'https://raw.githubusercontent.com/scikit-mobility/scikit-mobility/master/tutorial/data/geolife_sample.txt.gz'
    >>> df = pd.read_csv(url, sep=',', compression='gzip')
    >>> tdf = skmob.TrajDataFrame(df, latitude='lat', longitude='lon', user_id='user', datetime='datetime')
    >>> print(tdf.head())
             lat         lng            datetime  uid
    0  39.984094  116.319236 2008-10-23 05:53:05    1
    1  39.984198  116.319322 2008-10-23 05:53:06    1
    2  39.984224  116.319402 2008-10-23 05:53:11    1
    3  39.984211  116.319389 2008-10-23 05:53:16    1
    4  39.984217  116.319422 2008-10-23 05:53:21    1
    >>> # detect the stops first
    >>> stdf = detection.stops(tdf, stop_radius_factor=0.5, minutes_for_a_stop=20.0, spatial_radius_km=0.2, leaving_time=True)
    >>> # cluster the stops
    >>> cstdf = clustering.cluster(stdf, cluster_radius_km=0.1, min_samples=1)
    >>> print(cstdf.head())
             lat         lng            datetime  uid    leaving_datetime  cluster
    0  39.978030  116.327481 2008-10-23 06:01:37    1 2008-10-23 10:32:53        0
    1  40.013820  116.306532 2008-10-23 11:10:19    1 2008-10-23 23:45:27        1
    2  39.978419  116.326870 2008-10-24 00:21:52    1 2008-10-24 01:47:30        0
    3  39.981166  116.308475 2008-10-24 02:02:31    1 2008-10-24 02:30:29       42
    4  39.981431  116.309902 2008-10-24 02:30:29    1 2008-10-24 03:16:35       41    
    >>> print(cstdf.parameters)
    {'detect': {'function': 'stops', 'stop_radius_factor': 0.5, 'minutes_for_a_stop': 20.0, 'spatial_radius_km': 0.2, 'leaving_time': True, 'no_data_for_minutes': 1000000000000.0, 'min_speed_kmh': None}, 'cluster': {'function': 'cluster', 'cluster_radius_km': 0.1, 'min_samples': 1}}

    References
    ----------
    .. [DBSCAN] DBSCAN implementation, scikit-learn, https://scikit-learn.org/stable/modules/generated/sklearn.cluster.DBSCAN.html
    .. [RT2004] Ramaswamy, H. & Toyama, K. (2004) Project Lachesis: parsing and modeling location histories. In International Conference on Geographic Information Science, 106-124, http://kentarotoyama.com/papers/Hariharan_2004_Project_Lachesis.pdf
    """
    # Sort
    tdf = tdf.sort_by_uid_and_datetime()

    # Save function arguments and values in a dictionary
    frame = inspect.currentframe()
    args, _, _, arg_values = inspect.getargvalues(frame)
    arguments = dict([('function', cluster.__name__)]+[(i, arg_values[i]) for i in args[1:]])

    groupby = []

    if utils.is_multi_user(tdf):
        groupby.append(constants.UID)
    # if utils.is_multi_trajectory(data):
    #     groupby.append(constants.TID)

    stops_df = tdf
    # stops_df = detection.stops(data, stop_radius_factor=0.5, \
    #         minutes_for_a_stop=20.0, spatial_radius=0.2, leaving_time=True)

    if len(groupby) > 0:
        # Apply cluster stops to each group of points
        ctdf = stops_df.groupby(groupby, group_keys=False, as_index=False).apply(_cluster_trajectory,
                                cluster_radius_km=cluster_radius_km, min_samples=min_samples).reset_index(drop=True)
    else:
        ctdf = _cluster_trajectory(stops_df, cluster_radius_km=cluster_radius_km, min_samples=min_samples).reset_index(drop=True)

    # TODO: remove the following line when issue #71 (Preserve the TrajDataFrame index during preprocessing operations) is solved.
    ctdf.reset_index(inplace=True, drop=True)

    ctdf.parameters = tdf.parameters
    ctdf.set_parameter(constants.CLUSTERING_PARAMS, arguments)
    return ctdf


def _cluster_trajectory(tdf, cluster_radius_km, min_samples):
    # From dataframe convert to numpy matrix
    lat_lng_dtime_other = list(utils.to_matrix(tdf))
    columns_order = list(tdf.columns)

    l2x, cluster_IDs = _cluster_array(lat_lng_dtime_other, cluster_radius_km, min_samples)

    clusters_df = nparray_to_trajdataframe(lat_lng_dtime_other, utils.get_columns(tdf), {})
    # Put back to the original order
    clusters_df = clusters_df[columns_order]

    clusters_df.loc[:, 'cluster'] = cluster_IDs

    return clusters_df


def group_by_label(X, labels):
    """
     return a dictionary 'l2x' in which the elements 'x' of list 'X'
     are grouped according to 'labels'
    """
    l2x = dict([(l ,[]) for l in set(labels)])
    for x ,l in list(zip(X ,labels)):
        l2x[l] += [x]
    return l2x


def _cluster_array(lat_lng_dtime_other, cluster_radius_km, min_samples, verbose=False):

    X = np.array([[point[0], point[1]] for point in lat_lng_dtime_other])

    # Compute DBSCAN
    eps_rad = cluster_radius_km / kms_per_radian

    db = DBSCAN(eps=eps_rad, min_samples=min_samples, algorithm='ball_tree', metric='haversine')
    clus = db.fit(np.radians(X))
    # core_samples = clus.core_sample_indices_
    labels = clus.labels_

    # Number of clusters in labels, ignoring noise if present.
    n_clusters_ = len(set(labels)) - (1 if -1 in labels else 0)

    if verbose:
        print('Estimated number of clusters: %d' % n_clusters_)
    l02x = group_by_label(X, labels)

    # Map cluster index to most frequent location: label2fml
    c2mfl = dict([(c[1] ,i) for i ,c in \
                  enumerate(sorted([[len(v) ,l] for l ,v in l02x.items() if l> -0.5], reverse=True))])
    l2x = dict([(c2mfl[k], v) for k, v in l02x.items() if k > -0.5])
    try:
        l2x[-1] = l02x[-1.]
    except KeyError:
        pass

    return l2x, [c2mfl[k] for k in labels]



from ..utils import constants, gislib
import pandas as pd
import geopandas as gpd
import shapely
import os
import errno
import requests
import numpy as np

distance = gislib.getDistance


def diff_seconds(t_0, t_1):
    return (t_1 - t_0).total_seconds()


def is_multi_user(data):
    if constants.UID in data.columns:
        return True

    return False


def is_multi_trajectory(data):
    if constants.TID in data.columns:
        return True

    return False


def to_matrix(data, columns=None):
    if columns is None:
        columns = [constants.LATITUDE, constants.LONGITUDE, constants.DATETIME]
        columns = columns + list(set(data.columns) - set(columns))

    return data[columns].values


def get_columns(data):
    columns = [constants.LATITUDE, constants.LONGITUDE, constants.DATETIME]
    columns = columns + list(set(data.columns) - set(columns))

    return columns


def to_dataframe(data, columns):
    # Reorder columns to maintain the original order
    df = pd.DataFrame(data, columns=columns)

    return df


def assign_crs(shape, crs):
    if crs is not None:
        return shape.to_crs(crs)

    return shape


def to_geodataframe(df, keep=False, latitude=constants.LATITUDE, longitude=constants.LONGITUDE,
                    crs=constants.DEFAULT_CRS):
    geometry = [shapely.geometry.Point(xy) for xy in zip(df[longitude], df[latitude])]

    gdf = gpd.GeoDataFrame(df, crs=crs, geometry=geometry)
    df.drop('geometry', inplace=True, axis=1)

    if keep is False:
        return gdf.drop([longitude, latitude], axis=1)

    return gdf


def silentremove(filename):
    try:

        os.remove(filename)

    except OSError as e:  # this would be "except OSError, e:" before Python 2.6

        if e.errno != errno.ENOENT:  # errno.ENOENT = no such file or directory
            raise  # re-raise exception if a different error occurred


def setattrpandas(df, name, value):
    df._metadata.append(name)
    df.name = value

    return df


def group_df_by_time(tdf, freq_str='1D', offset_value=0, offset_unit='hours', add_starting_location=False,
                     dtformat='%Y-%m-%d %H:%M:%S'):
    """
    Split a `TrajDataFrame` into subtrajectories of fixed temporal length (`freq_str`).

    :param tdf: TrajDataFrame
        `TrajDataFrame` to split.

    :param freq_str: str
        `freq` parameter of `pd.date_range`. default: '1D' (one day).

    :param offset_value: float
        value of the offset time used to shift the start time of each subtrajectories.

    :param offset_unit: str
        time unit of the offset time used to shift the start time of each subtrajectories.

    :param add_starting_location: bool
        if `True`, the last point of the previous subtrajectory will be appended
         at the beginning of the next subtrajectory.

    :param dtformat: str
        datetime format.

    :return: list containing the subtrajectories

    """
    df = tdf.sort_values([constants.DATETIME])

    offset = pd.Timedelta(offset_value, offset_unit)  # datetime.timedelta(seconds=offset_secs)
    t_init = pd.to_datetime(df[constants.DATETIME].min().date())
    t_end = pd.to_datetime(df[constants.DATETIME].max().date()) + pd.Timedelta(days=1)  # datetime.timedelta(days=1)

    rg = pd.date_range((t_init + offset).strftime(dtformat), end=(t_end + offset).strftime(dtformat), freq=freq_str)

    groups = []
    for st, en in list(zip(rg, rg[1:])):
        dfslice = df.loc[(df[constants.DATETIME] >= st.strftime(dtformat)) &
                         (df[constants.DATETIME] < en.strftime(dtformat))]
        if len(dfslice) > 0:
            #             groups[st.date()] = [dfslice]
            groups += [dfslice.reset_index(drop=True)]

    if add_starting_location:
        for i in range(1, len(groups)):
            groups[i] = pd.concat([groups[i - 1][-1:], groups[i]]).reset_index(drop=True)

    return groups


def frequency_vector(trajectory):
    freq = trajectory.groupby([constants.UID, constants.LATITUDE, constants.LONGITUDE]).size().reset_index(
        name=constants.FREQUENCY)
    return freq.sort_values(by=[constants.UID, constants.FREQUENCY])


def probability_vector(trajectory):
    freq = trajectory.groupby([constants.UID, constants.LATITUDE, constants.LONGITUDE]).size().reset_index(
        name=constants.FREQUENCY)
    prob = pd.merge(freq, trajectory.groupby(constants.UID).size().reset_index(name=constants.TOTAL_FREQ),
                    left_on=constants.UID, right_on=constants.UID)
    prob[constants.PROBABILITY] = prob[constants.FREQUENCY] / prob[constants.TOTAL_FREQ]
    return prob[[constants.UID, constants.LATITUDE, constants.LONGITUDE, constants.PROBABILITY]].sort_values(
        by=[constants.UID, constants.PROBABILITY])


def date_time_precision(dt, precision):
    result = ""
    if precision == "Year" or precision == "year":
        result += str(dt.year)
    elif precision == "Month" or precision == "month":
        result += str(dt.year) + str(dt.month)
    elif precision == "Day" or precision == "day":
        result += str(dt.year) + str(dt.month) + str(dt.day)
    elif precision == "Hour" or precision == "hour":
        result += str(dt.year) + str(dt.month) + str(dt.day) + str(dt.month)
    elif precision == "Minute" or precision == "minute":
        result += str(dt.year) + str(dt.month) + str(dt.day) + str(dt.month) + str(dt.minute)
    elif precision == "Second" or precision == "second":
        result += str(dt.year) + str(dt.month) + str(dt.day) + str(dt.month) + str(dt.minute) + str(dt.second)
    return result


def bbox_from_points(points, crs=None):
    coords = points.total_bounds

    base = shapely.geometry.box(coords[0], coords[1], coords[2], coords[3], ccw=True)
    base = gpd.GeoDataFrame(geometry=[base], crs=constants.DEFAULT_CRS)

    if crs is None:
        return base

    return base.to_crs(crs)


def bbox_from_area(area, bbox_side_len=500, crs=None):
    centroid = area.iloc[0].geometry.centroid

    # get North-East corner
    ne = [float(coord) + (bbox_side_len / 2) for coord in centroid]

    # get South-West corner
    sw = [float(coord) - (bbox_side_len / 2) for coord in centroid]

    # build bbox from NE,SW corners
    bbox = shapely.geometry.box(sw[0], sw[1], ne[0], ne[1], ccw=True)

    base = gpd.GeoDataFrame(geometry=[bbox], crs=constants.DEFAULT_CRS)

    if crs is None:
        return base

    return base.to_crs(crs)


def bbox_from_name(query, which_result=0, crs=None):
    """
    Create a GeoDataFrame from a single place name query.
    (adapted from https://github.com/gboeing/osmnx)

    Parameters
    ----------
    query : string or dict
        query string or structured query dict to geocode/download

    which_result : int
        number of result to return (`which_result=-1` to return all results)

    Returns
    -------
    GeoDataFrame

    Example
    -------

    gdf = gdf_from_string("Florence, Italy")

    """
    nominatim_url = "https://nominatim.openstreetmap.org/search.php?" + \
                    "q=%s&polygon_geojson=1&format=json" % query

    response = requests.get(nominatim_url)
    data = response.json()

    if len(data) > 0:

        features = []
        for result in data:
            # extract data elements from the JSON response
            bbox_south, bbox_north, bbox_west, bbox_east = [float(x) for x in result['boundingbox']]

            coords = result['geojson']['coordinates']
            try:
                # it is a MultiPolygon
                geometry = shapely.geometry.MultiPolygon( \
                    [shapely.geometry.Polygon(c[0], [inner for inner in c[1:]]) for c in coords])
            except TypeError:
                try:
                    # it is a Polygon
                    geometry = shapely.geometry.MultiPolygon( \
                        [shapely.geometry.Polygon(coords[0], [inner for inner in coords[1:]])])
                except TypeError:
                    # it is something else
                    geometry = result['geojson']

            place = result['display_name']
            features += [{'type': 'Feature',
                          'geometry': geometry,
                          'properties': {'place_name': place,
                                         'bbox_north': bbox_north,
                                         'bbox_south': bbox_south,
                                         'bbox_east': bbox_east,
                                         'bbox_west': bbox_west}}]

        gdf = gpd.GeoDataFrame.from_features(features)

        if crs is None:
            gdf.crs = constants.DEFAULT_CRS
        else:
            gdf = gdf.to_crs(crs)

        if which_result >= 0:
            gdf = gdf.loc[[which_result]]

    else:
        gdf = gpd.GeoDataFrame()

    return gdf


def nearest(origin, tessellation, col):
    """
    Brute force approach to find, for each point in a geodataframe, the nearest point into another geodataframe. It
    returns a Pandas Series containing the value in col for each matching row.
    :param origin: GeoDataFrame
    :param tessellation: GeoDataFrame
    :param col: column containing the value to return from the tessellation
    :return: Series
    """

    def _nearest(df, points):

        near = float("+inf")
        point = None

        main = (df['geometry'].y, df['geometry'].x)

        for index, row in points.iterrows():

            p = row['geometry']
            d = distance(main, (p.y, p.x))

            if d < near:
                near = d
                point = index

        return point

    return tessellation.iloc[origin.apply(_nearest, args=(tessellation,), axis=1)][col]


def get_geom_centroid(geom, return_lat_lng=False):
    """
    Compute the centroid of a Polygon or Multipolygon.

    :param geom: shapely Polygon or Multipolygon
        geometry, Polygon or Multipolygon, whose centroid will be computed.

    :param return_lat_lng: bool
        if `True`, the first coordinate in the returned list is the centroid's latitude, otherwise it is the longitude.

    :return: list
        coordinates of `geom`'s centroid.

    """
    try:
        lonO, latO = np.array(geom.centroid.xy).flatten()
    except:
        if type(geom) == shapely.geometry.multipolygon.MultiPolygon:
            m_n = [[np.mean(pol.exterior.xy, axis=1), len(pol.exterior.xy[0])] for pol in geom]
            lonO, latO = np.sum([[ln*n, la*n] for (ln,la),n in m_n], axis=0) / np.sum(m_n, axis=0)[1]
        elif type(geom) == shapely.geometry.polygon.Polygon:
            lonO, latO = np.mean(geom.exterior.xy, axis=1)
        else:
            lonO, latO = np.mean(geom.xy, axis=1)
    if return_lat_lng:
        return [latO, lonO]
    else:
        return [lonO, latO]




from ..utils import constants, utils
import folium
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import shapely
from geojson import LineString
import geopandas as gpd
import json


COLOR = {
    0: '#FF0000',  # Red
    1: '#008000',  # Green
    2: '#000080',  # Navy
    3: '#800000',  # Maroon
    4: '#FFD700',  # Gold
    5: '#00FF00',  # Lime
    6: '#800080',  # Purple
    7: '#00FFFF',  # Aqua
    8: '#DC143C',  # Crimson
    9: '#0000FF',  # Blue
    10: '#F08080',  # LightCoral
    11: '#FF00FF',  # Fuchsia
    12: '#FF8C00',  # DarkOrange
    13: '#6A5ACD',  # SlateBlue
    14: '#8B4513',  # SaddleBrown
    15: '#1E90FF',  # DodgerBlue
    16: '#FFFF00',  # Yellow
    17: '#808080',  # Gray
    18: '#008080',  # Teal
    19: '#9370DB',  # MediumPurple
    20: '#2F4F4F'  # DarkSlateGray
}


def get_color(k=-1, color_dict=COLOR):
    """
    Return a color (random if "k" is negative)
    """
    if k < 0:
        return np.random.choice(list(color_dict.values()))  # color_dict[random.randint(0,20)]
    else:
        return color_dict[k % 21]


def random_hex():
    r = lambda: np.random.randint(0,255)
    return '#%02X%02X%02X' % (r(),r(),r())


traj_style_function = lambda weight, color, opacity: \
    (lambda feature: dict(color=color, weight=weight, opacity=opacity))

def plot_trajectory(tdf, map_f=None, max_users=10, max_points=1000, style_function=traj_style_function,
                    tiles='cartodbpositron', zoom=12, hex_color=-1, weight=2, opacity=0.75, start_end_markers=True):
    """
    :param tdf: TrajDataFrame
         TrajDataFrame to be plotted.

    :param map_f: folium.Map
        `folium.Map` object where the trajectory will be plotted. If `None`, a new map will be created.

    :param max_users: int
        maximum number of users whose trajectories should be plotted.

    :param max_points: int
        maximum number of points per user to plot.
        If necessary, a user's trajectory will be down-sampled to have at most `max_points` points.

    :param style_function: lambda function
        function specifying the style (weight, color, opacity) of the GeoJson object.

    :param tiles: str
        folium's `tiles` parameter.

    :param zoom: int
        initial zoom.

    :param hex_color: str or int
        hex color of the trajectory line. If `-1` a random color will be generated for each trajectory.

    :param weight: float
        thickness of the trajectory line.

    :param opacity: float
        opacity (alpha level) of the trajectory line.

    :param start_end_markers: bool
        add markers on the start and end points of the trajectory.

    :return: `folium.Map` object with the plotted trajectories.

    """
    # group by user and keep only the first `max_users`
    nu = 0


    try:
        # column 'uid' is present in the TrajDataFrame
        groups = tdf.groupby(constants.UID)
    except KeyError:
        # column 'uid' is not present
        groups = [[None, tdf]]

    for user, df in groups:

        if nu >= max_users:
            break
        nu += 1

        traj = df[[constants.LONGITUDE, constants.LATITUDE]]

        if max_points is None:
            di = 1
        else:
            di = max(1, len(traj) // max_points)
        traj = traj[::di]

        if nu == 1 and map_f is None:
            # initialise map
            center = list(np.median(traj, axis=0)[::-1])
            map_f = folium.Map(location=center, zoom_start=zoom, tiles=tiles)

        trajlist = traj.values.tolist()
        line = LineString(trajlist)

        if hex_color == -1:
            color = get_color(hex_color)
        else:
            color = hex_color

        tgeojson = folium.GeoJson(line,
                                  name='tgeojson',
                                  style_function=style_function(weight, color, opacity)
                                  )
        tgeojson.add_to(map_f)

        if start_end_markers:

            dtime, la, lo = df.loc[df['datetime'].idxmin()]\
                [[constants.DATETIME, constants.LATITUDE, constants.LONGITUDE]].values
            dtime = pd.datetime.strftime(dtime, '%Y/%m/%d %H:%M')
            mker = folium.Marker(trajlist[0][::-1], icon=folium.Icon(color='green'))
            popup = folium.Popup('<i>Start</i><BR>{}<BR>Coord: <a href="https://www.google.co.uk/maps/place/{},{}" target="_blank">{}, {}</a>'.\
                          format(dtime, la, lo, np.round(la, 4), np.round(lo, 4)), max_width=300)
            mker = mker.add_child(popup)
            mker.add_to(map_f)

            dtime, la, lo = df.loc[df['datetime'].idxmax()]\
                [[constants.DATETIME, constants.LATITUDE, constants.LONGITUDE]].values
            dtime = pd.datetime.strftime(dtime, '%Y/%m/%d %H:%M')
            mker = folium.Marker(trajlist[-1][::-1], icon=folium.Icon(color='red'))
            popup = folium.Popup('<i>End</i><BR>{}<BR>Coord: <a href="https://www.google.co.uk/maps/place/{},{}" target="_blank">{}, {}</a>'.\
                          format(dtime, la, lo, np.round(la, 4), np.round(lo, 4)), max_width=300)
            mker = mker.add_child(popup)
            mker.add_to(map_f)

    return map_f


def plot_stops(stdf, map_f=None, max_users=10, tiles='cartodbpositron', zoom=12,
               hex_color=-1, opacity=0.3, radius=12, popup=True):
    """
    :param stdf: TrajDataFrame
         Requires a TrajDataFrame with stops or clusters, output of `preprocessing.detection.stops`
         or `preprocessing.clustering.cluster`. The column `constants.LEAVING_DATETIME` must be present.

    :param map_f: folium.Map
        `folium.Map` object where the stops will be plotted. If `None`, a new map will be created.

    :param max_users: int
        maximum number of users whose stops should be plotted.

    :param tiles: str
        folium's `tiles` parameter.

    :param zoom: int
        initial zoom.

    :param hex_color: str or int
        hex color of the stop markers. If `-1` a random color will be generated for each user.

    :param opacity: float
        opacity (alpha level) of the stop makers.

    :param radius: float
        size of the markers.

    :param popup: bool
        if `True`, when clicking on a marker a popup window displaying information on the stop will appear.

    :return: `folium.Map` object with the plotted stops.

    """
    if map_f is None:
        # initialise map
        lo_la = stdf[['lng', 'lat']].values
        center = list(np.median(lo_la, axis=0)[::-1])
        map_f = folium.Map(location=center, zoom_start=zoom, tiles=tiles)

    # group by user and keep only the first `max_users`
    nu = 0

    try:
        # column 'uid' is present in the TrajDataFrame
        groups = stdf.groupby(constants.UID)
    except KeyError:
        # column 'uid' is not present
        groups = [[None, stdf]]

    for user, df in groups:
        if nu >= max_users:
            break
        nu += 1

        if hex_color == -1:
            color = get_color(hex_color)
        else:
            color = hex_color

        for idx, row in df.iterrows():

            la = row[constants.LATITUDE]
            lo = row[constants.LONGITUDE]
            t0 = row[constants.DATETIME]
            t1 = row[constants.LEAVING_DATETIME]
            u = user
            try:
                ncluster = row[constants.CLUSTER]
                cl = '<BR>Cluster: {}'.format(ncluster)
                color = get_color(ncluster)
            except (KeyError, NameError):
                cl = ''

            fpoly = folium.RegularPolygonMarker([la, lo],
                                        radius=radius,
                                        color=color,
                                        fill_color=color,
                                        fill_opacity=opacity
                                        )
            if popup:
                popup = folium.Popup('User: {}<BR>Coord: <a href="https://www.google.co.uk/maps/place/{},{}" target="_blank">{}, {}</a><BR>Arr: {}<BR>Dep: {}{}' \
                    .format(u, la, lo, np.round(la, 4), np.round(lo, 4),
                            pd.datetime.strftime(t0, '%Y/%m/%d %H:%M'),
                            pd.datetime.strftime(t1, '%Y/%m/%d %H:%M'), cl), max_width=300)
                fpoly = fpoly.add_child(popup)

            fpoly.add_to(map_f)

    return map_f


def plot_diary(cstdf, user, start_datetime=None, end_datetime=None, ax=None, legend=False):
    """
        Plot a mobility diary of an individual in a TrajDataFrame. It requires a TrajDataFrame with clusters, output of `preprocessing.clustering.cluster`. The column `constants.CLUSTER` must be present.

        Parameters
        ----------
        user : str or int
            user identifier whose diary should be plotted.

        start_datetime : datetime.datetime, optional
            only stops made after this date will be plotted. If `None` the datetime of the oldest stop will be selected. The default is `None`.

        end_datetime : datetime.datetime, optional
            only stops made before this date will be plotted. If `None` the datetime of the newest stop will be selected. The default is `None`.

        ax : matplotlib.axes, optional
            axes where the diary will be plotted. If `None` a new ax is created. The default is `None`.

        legend : bool, optional
            If `True`, legend with cluster IDs is shown. The default is `False`.

        Returns
        -------
        matplotlib.axes
            the `matplotlib.axes` object of the plotted diary.

    """
    if ax is None:
        fig, ax = plt.subplots(figsize=(20, 2))

    if user is None:
        df = cstdf
    else:
        df = cstdf[cstdf[constants.UID] == user]

    # TODO: add warning if days between start_datetime and end_datetime do not overlap with cstdf
    if start_datetime is None:
        start_datetime = df[constants.DATETIME].min()
    if end_datetime is None:
        end_datetime = df[constants.LEAVING_DATETIME].max()

    current_labels = []

    for idx, row in df.iterrows():

        t0 = row[constants.DATETIME]
        t1 = row[constants.LEAVING_DATETIME]
        cl = row[constants.CLUSTER]

        color = get_color(cl)
        if start_datetime <= t0 <= end_datetime:
            if cl in current_labels:
                ax.axvspan(t0, t1, lw=0.0, alpha=0.75, color=color)
            else:
                current_labels += [cl]
                ax.axvspan(t0, t1, lw=0.0, alpha=0.75, color=color, label=cl)

    plt.xlim(start_datetime, end_datetime)

    if legend:
        handles, labels_str = ax.get_legend_handles_labels()
        labels = list(map(int, labels_str))
        # sort them by labels
        import operator
        hl = sorted(zip(handles, labels), key=operator.itemgetter(1))
        handles2, labels2 = zip(*hl)

        ax.legend(handles2, labels2, ncol=15, bbox_to_anchor=(1., -0.2), frameon=0)

    ax.set_title('user %s' % user)

    return ax



flow_style_function = lambda weight, color, opacity, weight_factor, flow_exp: \
    (lambda feature: dict(color=color, weight=weight_factor * weight ** flow_exp, opacity=opacity)) #, dashArray='5, 5'))


def plot_flows(fdf, map_f=None, min_flow=0, tiles='cartodbpositron', zoom=6, flow_color='red', opacity=0.5,
               flow_weight=5, flow_exp=0.5, style_function=flow_style_function,
               flow_popup=False, num_od_popup=5, tile_popup=True, radius_origin_point=5,
               color_origin_point='#3186cc'):
    """
    :param fdf: FlowDataFrame
        `FlowDataFrame` to visualize.

    :param map_f: folium.Map
        `folium.Map` object where the flows will be plotted. If `None`, a new map will be created.

    :param min_flow: float
        only flows larger than `min_flow` will be plotted.

    :param tiles: str
        folium's `tiles` parameter.

    :param zoom: int
        initial zoom.

    :param flow_color: str
        color of the flow edges

    :param opacity: float
        opacity (alpha level) of the flow edges.

    :param flow_weight: float
        weight factor used in the function to compute the thickness of the flow edges.

    :param flow_exp: float
        weight exponent used in the function to compute the thickness of the flow edges.

    :param style_function: lambda function
        GeoJson style function.

    :param flow_popup: bool
        if `True`, when clicking on a flow edge a popup window displaying information on the flow will appear.

    :param num_od_popup: int
        number of origin-destination pairs to show in the popup window of each origin location.

    :param tile_popup: bool
        if `True`, when clicking on a location marker a popup window displaying information on the flows
        departing from that location will appear.

    :param radius_origin_point: float
        size of the location markers.

    :param color_origin_point: str
        color of the location markers.

    :return: `folium.Map` object with the plotted flows.

    """
    if map_f is None:
        # initialise map
        lon, lat = np.mean(np.array(list(fdf.tessellation.geometry.apply(utils.get_geom_centroid).values)), axis=0)
        map_f = folium.Map(location=[lat,lon], tiles=tiles, zoom_start=zoom)

    mean_flows = fdf[constants.FLOW].mean()

    O_groups = fdf.groupby(by=constants.ORIGIN)
    for O, OD in O_groups:

        geom = fdf.get_geometry(O)
        lonO, latO = utils.get_geom_centroid(geom)

        for D, T in OD[[constants.DESTINATION, constants.FLOW]].values:
            if O == D:
                continue
            if T < min_flow:
                continue

            geom = fdf.get_geometry(D)
            lonD, latD = utils.get_geom_centroid(geom)

            gjc = LineString([(lonO,latO), (lonD,latD)])

            fgeojson = folium.GeoJson(gjc,
                                      name='geojson',
                                      style_function = style_function(T / mean_flows, flow_color, opacity,
                                                                      flow_weight, flow_exp)
                                      )
            if flow_popup:
                popup = folium.Popup('flow from %s to %s: %s'%(O, D, int(T)), max_width=300)
                fgeojson = fgeojson.add_child(popup)

            fgeojson.add_to(map_f)

    if radius_origin_point > 0:
        for O, OD in O_groups:

            name = 'origin: %s' % O.replace('\'', '_')
            T_D = [[T, D] for D, T in OD[[constants.DESTINATION, constants.FLOW]].values]
            trips_info = '<br/>'.join(["flow to %s: %s" %
                                       (dd.replace('\'', '_'), int(tt)) \
                                       for tt, dd in sorted(T_D, reverse=True)[:num_od_popup]])

            geom = fdf.get_geometry(O)
            lonO, latO = utils.get_geom_centroid(geom)
            fmarker = folium.CircleMarker([latO, lonO],
                                          radius=radius_origin_point,
                                          weight=2,
                                          color=color_origin_point,
                                          fill=True, fill_color=color_origin_point
                                          )
            if tile_popup:
                popup = folium.Popup(name+'<br/>'+trips_info, max_width=300)
                fmarker = fmarker.add_child(popup)
            fmarker.add_to(map_f)

    return map_f


default_style_func_args = {'weight': 1, 'color': 'random', 'opacity': 0.5, 'fillColor': 'red', 'fillOpacity': 0.25}

geojson_style_function = lambda weight, color, opacity, fillColor, fillOpacity: \
    (lambda feature: dict(weight=weight, color=color, opacity=opacity, fillColor=fillColor, fillOpacity=fillOpacity))


def add_to_map(gway, g, map_osm, style_func_args, popup_features=[]):

    styles = []
    for k in ['weight', 'color', 'opacity', 'fillColor', 'fillOpacity']:
        if k in style_func_args:
            if callable(style_func_args[k]):
                styles += [style_func_args[k](g)]
            else:
                styles += [style_func_args[k]]
        else:
            styles += [default_style_func_args[k]]
    weight, color, opacity, fillColor, fillOpacity = styles


    if type(gway) == shapely.geometry.multipolygon.MultiPolygon:

        # Multipolygon
        for gg in gway:
            if color == 'random':
                color = random_hex()
                fillColor = color

            vertices = list(zip(*gg.exterior.xy))
            gj = folium.GeoJson({"type": "Polygon", "coordinates": [vertices]},
                                style_function=geojson_style_function(weight=weight, color=color, opacity=opacity,
                                                                      fillColor=fillColor, fillOpacity=fillOpacity))

    elif type(gway) == shapely.geometry.polygon.Polygon:

        # Polygon
        if color == 'random':
            color = random_hex()
            fillColor = color

        vertices = list(zip(*gway.exterior.xy))
        gj = folium.GeoJson({"type": "Polygon", "coordinates": [vertices]},
                            style_function=geojson_style_function(weight=weight, color=color, opacity=opacity,
                                                                  fillColor=fillColor, fillOpacity=fillOpacity))

    elif type(gway) == shapely.geometry.multilinestring.MultiLineString:

        # MultiLine
        for gg in gway:
            if color == 'random':
                color = random_hex()
                fillColor = color

            vertices = list(zip(*gg.xy))
            gj = folium.GeoJson({"type": "LineString", "coordinates": vertices},
                                style_function=geojson_style_function(weight=weight, color=color, opacity=opacity,
                                                                      fillColor=fillColor, fillOpacity=fillOpacity))

    elif type(gway) == shapely.geometry.linestring.LineString:

        # LineString
        if color == 'random':
            color = random_hex()
            fillColor = color
        vertices = list(zip(*gway.xy))
        gj = folium.GeoJson({"type": "LineString", "coordinates": vertices},
                            style_function=geojson_style_function(weight=weight, color=color, opacity=opacity,
                                                                  fillColor=fillColor, fillOpacity=fillOpacity))

    else:

        # Point
        if color == 'random':
            color = random_hex()
            fillColor = color

        point = list(zip(*gway.xy))[0]
        #         gj = folium.CircleMarker(
        gj = folium.Circle(
            location=point[::-1],
            radius=5,
            color=color,  # '#3186cc',
            fill=True,
            fill_color=fillColor
        )

    popup = []
    for pf in popup_features:
        try:
            popup += ['%s: %s' % (pf, g[pf])]
        except KeyError:
            pass

    try:
        popup = '<br>'.join(popup)
        popup += json.dumps(g.tags)
        popup = popup.replace("""'""", """_""")
    except AttributeError:
        pass
    if len(popup) > 0:
        gj.add_child(folium.Popup(popup, max_width=300))

    gj.add_to(map_osm)

    return map_osm


def plot_gdf(gdf, map_osm=None, maxitems=-1, style_func_args={}, popup_features=[],
            tiles='cartodbpositron', zoom=6, geom_col='geometry'):
    """
    :param gdf: GeoDataFrame
        GeoDataFrame to visualize.

    :param map_osm: folium.Map
        `folium.Map` object where the GeoDataFrame `gdf` will be plotted. If `None`, a new map will be created.

    :param maxitems: int
        maximum number of tiles to plot. If `-1`, all tiles will be plotted.

    :param style_func_args: dict
        dictionary to pass the following style parameters (keys) to the GeoJson style function of the polygons:
        'weight', 'color', 'opacity', 'fillColor', 'fillOpacity'

    :param popup_features: list
        when clicking on a tile polygon, a popup window displaying the information in the
        columns of `gdf` listed in `popup_features` will appear.

    :param tiles: str
        folium's `tiles` parameter.

    :param zoom: int
        initial zoom.

    :param geom_col: str
         name of the geometry column of `gdf`.

    :return: `folium.Map` object with the plotted GeoDataFrame.

    """
    if map_osm is None:
        # initialise map
        lon, lat = np.mean(np.array(list(gdf[geom_col].apply(utils.get_geom_centroid).values)), axis=0)
        map_osm = folium.Map(location=[lat, lon], tiles=tiles, zoom_start=zoom)

    count = 0
    for k in gdf.index:
        g = gdf.loc[k]

        if type(g[geom_col]) == gpd.geoseries.GeoSeries:
            for i in range(len(g[geom_col])):
                map_osm = add_to_map(g[geom_col].iloc[i], g.iloc[i], map_osm,
                                     popup_features=popup_features,
                                     style_func_args=style_func_args)
        else:
            map_osm = add_to_map(g[geom_col], g, map_osm,
                                 popup_features=popup_features,
                                 style_func_args=style_func_args)

        count += 1
        if count == maxitems:
            break

    return map_osm



import numpy as np
from scipy import stats
from sklearn.metrics import r2_score, mean_squared_error


def common_part_of_commuters(values1, values2):
    """
    Compute the common part of commuters for two pairs of fluxes.

    :param values1: the values for the first array
    :type values1: numpy array

    :param values2: the values for the second array
    :type values1: numpy array

    :return: float
        the common part of commuters
    """
    return 2.0 * np.sum(np.minimum(values1, values2)) / (np.sum(values2) + np.sum(values2))


def common_part_of_links(values1, values2):
    """
    Compute the common part of commuters for two pairs of fluxes.

    :param values1: the values for the first array
    :type values1: numpy array

    :param values2: the values for the second array
    :type values2: numpy array

    :return: float
        the common part of commuters
    """

    def check_condition(value):
        if value > 0:
            return 1
        return 0

    cpl_num = 0.0
    cpl_den = 0.0
    for val1, val2 in zip(values1, values2):
        cpl_num += check_condition(val1) * check_condition(val2, 1)
        cpl_den += check_condition(val1) + check_condition(val2)
    return 2.0 * cpl_num / cpl_den


def common_part_of_commuters_distance(values1, values2):
    """
    Compute the common part of commuters according to the distance.

    :param values1: the values for the first array
    :type values1: numpy array

    :param values2: the values for the second array
    :type values2: numpy array

    :return: float
        the common part of commuters according to the distance
    """
    max_val = max(max(values1), max(values2))
    bins = np.arange(0, max_val, 2)
    hist1, bin_edges1 = np.histogram(values1, bins)
    hist2, bin_edges2 = np.histogram(values2, bins)

    N, cpcd = sum(values1), 0.0
    for k1, k2 in zip(hist1, hist2):
        cpcd += min(k1, k2)
    cpcd /= N
    return cpcd


def r_squared(true, pred):
    """
    R^2 (coefficient of determination) regression score function.

    Best possible score is 1.0 and it can be negative
    (because the model can be arbitrarily worse).
    A constant model that always predicts the expected value of y,
    disregarding the input features, would get a R^2 score of 0.0.

    :param true: Ground truth (correct) target values.
    :type true: numpy array array-like of shape = (n_samples) or (n_samples, n_outputs)

    :param pred: Estimated target values.
    :type pred: numpy array-like of shape = (n_samples) or (n_samples, n_outputs)

    :return: float
    """
    return r2_score(true, pred)


def mse(true, pred):
    """
    Mean squared error regression loss

    :param true: Ground truth (correct) target values.
    :type true: numpy array array-like of shape = (n_samples) or (n_samples, n_outputs)

    :param pred: Estimated target values.
    :type pred: numpy array-like of shape = (n_samples) or (n_samples, n_outputs)

    :return: float
        A non-negative floating point value (the best value is 0.0)
    """
    return mean_squared_error(true, pred)


def rmse(true, pred):
    """
    Root mean squared error regression loss

    :param true: Ground truth (correct) target values.
    :type true: numpy array array-like of shape = (n_samples) or (n_samples, n_outputs)

    :param pred: Estimated target values.
    :type pred: numpy array-like of shape = (n_samples) or (n_samples, n_outputs)

    :return: float
        A non-negative floating point value (the best value is 0.0)
    """
    return np.sqrt(mse(true, pred))


def nrmse(true, pred):
    """
    Normalized mean squared error regression loss

    :param true: Ground truth (correct) target values.
    :type true: numpy array array-like of shape = (n_samples) or (n_samples, n_outputs)

    :param pred: Estimated target values.
    :type pred: numpy array-like of shape = (n_samples) or (n_samples, n_outputs)

    :return: float
        A non-negative floating point value (the best value is 0.0)
    """
    return rmse(true, pred) / np.sum(true)


def information_gain(true, pred):
    """
    The information gain

    :param true: Ground truth (correct) target values.
    :type true: numpy array array-like of shape = (n_samples) or (n_samples, n_outputs)

    :param pred: Estimated target values.
    :type pred: numpy array-like of shape = (n_samples) or (n_samples, n_outputs)

    :return: float
        A non-negative floating point value (the best value is 0.0)
    """
    N = np.sum(true)
    information_gain = 0.0
    for true_value, pred_value in zip(true, pred):
        information_gain += (1.0 * true_value / N) * np.log(true_value / pred_value)
    return information_gain


def pearson_correlation(true, pred):
    """
    Calculates a Pearson correlation coefficient and the p-value for testing non-correlation.
    The Pearson correlation coefficient measures the linear relationship between two datasets.
    Strictly speaking, Pearson’s correlation requires that each dataset be normally distributed. Like other correlation coefficients, this one varies between -1 and +1 with 0 implying no correlation. Correlations of -1 or +1 imply an exact linear relationship. Positive correlations imply that as x increases, so does y.
    Negative correlations imply that as x increases, y decreases.
    The p-value roughly indicates the probability of an uncorrelated system producing datasets that have a Pearson correlation at least as extreme as the one computed from these datasets. The p-values are not entirely reliable but are probably reasonable for datasets larger than 500 or so.

    :param true: Ground truth (correct) target values.
    :type true: numpy array array-like of shape = (n_samples) or (n_samples, n_outputs)

    :param pred: Estimated target values.
    :type pred: numpy array-like of shape = (n_samples) or (n_samples, n_outputs)

    :return: tuple
        (Pearson’s correlation coefficient, 2-tailed p-value)
    """
    return stats.pearsonr(true, pred)


def spearman_correlation(true, pred):
    """
    Calculates a Spearman rank-order correlation coefficient and the p-value to test for non-correlation.
    The Spearman correlation is a nonparametric measure of the monotonicity of the relationship between two datasets.
    Unlike the Pearson correlation, the Spearman correlation does not assume that both datasets are normally distributed.
    Like other correlation coefficients, this one varies between -1 and +1 with 0 implying no correlation.
    Correlations of -1 or +1 imply an exact monotonic relationship. Positive correlations imply that as x increases, so does y.
    Negative correlations imply that as x increases, y decreases. The p-value roughly indicates the probability of an uncorrelated system
    producing datasets that have a Spearman correlation at least as extreme as the one computed from these datasets.
    The p-values are not entirely reliable but are probably reasonable for datasets larger than 500 or so.

    :param true: Ground truth (correct) target values.
    :type true: numpy array array-like of shape = (n_samples) or (n_samples, n_outputs)

    :param pred: Estimated target values.
    :type pred: numpy array-like of shape = (n_samples) or (n_samples, n_outputs)

    :return: tuple
        (Spearman’s correlation coefficient, 2-tailed p-value)
    """
    return stats.spearmanr(true, pred)


def kullback_leibler_divergence(true, pred):
    """
    Compute the Kullback-Leibler divergence S = sum(pk * log(pk / qk), axis=0).

    :param true: Ground truth (correct) target values.
    :type true: numpy array array-like of shape = (n_samples) or (n_samples, n_outputs)

    :param pred: Estimated target values.
    :type pred: numpy array-like of shape = (n_samples) or (n_samples, n_outputs)

    :return: float
        the calculated Kullback-Leibler divergence
    """
    return stats.entropy(true, pred)


def max_error(true, pred):
    """
    The maximum error between the two arrays

    :param true: Ground truth (correct) target values.
    :type true: numpy array array-like of shape = (n_samples) or (n_samples, n_outputs)

    :param pred: Estimated target values.
    :type pred: numpy array-like of shape = (n_samples) or (n_samples, n_outputs)

    :return: float
        max error between the two samples
    """
    return np.max(np.subtract(true, pred))



"""
Useful constants
"""

UNIVERSAL_CRS = {"init": "epsg:3857"}
DEFAULT_CRS = {"init": "epsg:4326"}

UID = 'uid'
TID = 'tid'
LATITUDE = 'lat'
LONGITUDE = 'lng'
ORIGIN = 'origin'
DESTINATION = 'destination'
ORIGIN_LAT = 'origin_lat'
ORIGIN_LNG = 'origin_lng'
DESTINATION_LAT = 'destination_lat'
DESTINATION_LNG = 'destination_lng'
DATETIME = 'datetime'
FLOW = 'flow'
TILE_ID = 'tile_ID'
CLUSTER = 'cluster'
LEAVING_DATETIME = 'leaving_datetime'
FREQUENCY = "freq"
PROBABILITY = "prob"
TOTAL_FREQ = "T_freq"
COUNT = "count"
TEMP = "tmp"
PROPORTION = "prop"
TOT_OUTFLOW = "tot_outflow"
PRECISION_LEVELS = ["Year", "Month", "Day", "Hour", "Minute", "Second", "year", "month", "day", "hour", "minute",
                    "second"]

PRIVACY_RISK = "risk"
INSTANCE = "instance"
INSTANCE_ELEMENT = "instance_elem"
REIDENTIFICATION_PROBABILITY = "reid_prob"
RELEVANCE = 'relevance'

# PROPERTIES KEY
CRS = 'crs'
FILTERING_PARAMS = 'filter'
COMPRESSION_PARAMS = 'compress'
CLUSTERING_PARAMS = 'cluster'
DETECTION_PARAMS = 'detect'





from setuptools import setup
import sys
if sys.version_info >= (3,8):
    sys.exit('Sorry, Python > 3.7 is not supported')
from setuptools import setup, find_packages

long_description = open('README.md').read()


DEPENDENCIES = [
    'numpy',
    'pandas',
    'geopandas',
    'scipy',
    'powerlaw',
    'tqdm',
    'requests',
    'scikit-learn',
    'statsmodels',
    'folium',
    'matplotlib',
    'geojson',
    'shapely',
    'fiona'
]


TEST_DEPENDENCIES = [
    'hypothesis',
    'mock',
    'python-Levenshtein',
]


setup(
    name='scikit-mobility-unofficial',
    version='1.0',
    packages=['skmob', 'skmob.core', 'skmob.utils', 'skmob.io', 'skmob.measures', 'skmob.models', 'skmob.preprocessing', 'skmob.privacy', 'skmob.tessellation' ],
    #TODO: fix it with find_packages(include=["skmob", "skmob.*"])
    license='MIT',
    python_requires='>=3.6',
    description='A toolbox for analyzing and processing mobility data.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    maintainer='skmob Developers',
    maintainer_email='gianni.barlacchi@gmail.com',
    classifiers=['Intended Audience :: Science/Research',
                 'Intended Audience :: Developers',
                 'License :: OSI Approved',
                 'Programming Language :: Python',
                 'Topic :: Software Development',
                 'Topic :: Scientific/Engineering',
                 'Operating System :: Microsoft :: Windows',
                 'Operating System :: Unix',
                 'Operating System :: MacOS',
                 #'Programming Language :: Python :: 3',
                 #'Programming Language :: Python :: 3.5',
                 'Programming Language :: Python :: 3.6',
                 'Programming Language :: Python :: 3.7',
                 ],
    install_requires=DEPENDENCIES
    #extra_requires={'geometry_support':'geopandas'}
    )



import pandas as pd
import fiona
import json
import os
import fnmatch
from ..core.trajectorydataframe import TrajDataFrame
from ..core.flowdataframe import FlowDataFrame
from ..preprocessing import filtering, compression


def write(skmob_df, file):
    """
    Write a TrajDataFrame to a json file.

    :param skmob_df: TrajDataFrame
        object that will be saved.

    :param file: str
        path and name of the `json` output file.

    :return: None

    """
    datetime_columns = [c for c in skmob_df.columns
                        if pd.core.dtypes.common.is_datetime64_any_dtype(skmob_df[c].dtype)]

    with open(file, 'w') as f:
        json.dump([str(type(skmob_df)),
                   skmob_df.to_json(orient='split', date_unit='s'),
                   skmob_df.parameters,
                   datetime_columns], f)


def read(file):
    """
    Read a TrajDataFrame from a json file.

    :param file: str
        path and name of the `json` file to read.

    :return:
        object loaded from file.

    """
    with open(file, 'r') as f:
        df_type, json_df, parameters, datetime_columns = json.load(f)

    if 'TrajDataFrame' in df_type:
        tdf = TrajDataFrame(pd.read_json(json_df, orient='split', date_unit='s',
                                         convert_dates=datetime_columns), parameters=parameters)
        return tdf
    elif 'FlowDataFrame' in df_type:
        return None
    else:
        print('DataFrame type not recognised.')
        return None


def load_gpx(file, user_id=None):
    """
    :param str file: str
        path to the gpx file

    :param user_id: str or int
        name or ID of the user

    :return: a TrajDataFrame containing the trajectory
    :rtype: TrajDataFrame

    """
    track = [[p['properties']['time']] + list(p['geometry']['coordinates'])
             for p in fiona.open(file, layer='track_points')]

    tdf = TrajDataFrame(track, datetime=0, longitude=1, latitude=2)

    if user_id is not None:
        tdf['uid'] = [user_id for _ in range(len(tdf))]

    return tdf.sort_by_uid_and_datetime()


def load_geolife_trajectories(path_to_geolife_data_dir, user_ids=[],
                              filter_kwargs={'max_speed_kmh': 400},
                              compress_kwargs={'spatial_radius_km': 0.2}):
    """
    Load the Geolife trajectories in a TrajDataFrame

    :param path_to_geolife_data_dir: str
        local path of the directory 'Geolife Trajectories 1.3/'

    :param user_ids: list
        list of user IDs to load. If empty all users are loaded.

    :param filter_kwargs: dict
        arguments of `preprocessing.filtering.filter()`. If empty, data is not filtered.

    :param compress_kwargs: dict
        arguments of `preprocessing.compression.compress()`. If empty, data is not compressed.

    :return: TrajDataFrame
        a TrajDataFrame containing all trajectories
    :rtype: TrajDataFrame

    """
    tdf = pd.DataFrame()

    path = path_to_geolife_data_dir + 'Data/'
    if len(user_ids) == 0:
        user_ids = os.listdir(path)

    for uid in user_ids:

        try:
            all_files = fnmatch.filter(os.listdir(path + '%s/Trajectory/' % uid), "*.plt")

        except NotADirectoryError:
            continue

        dfg = (pd.read_csv(path + '%s/Trajectory/' % uid + f,
                           skiprows=6, header=None, usecols=[0, 1, 5, 6]) for f in all_files)
        df = pd.concat(dfg, ignore_index=True)
        df['datetime'] = df[5] + ' ' + df[6]
        df.drop(columns=[5, 6], inplace=True)
        df['uid'] = [str(uid) for _ in range(len(df))]

        tdf0 = TrajDataFrame(df, latitude=0, longitude=1)

        if len(filter_kwargs) > 0:
            tdf0 = filtering.filter(tdf0, **filter_kwargs)

        if len(compress_kwargs) > 0:
            tdf0 = compression.compress(tdf0, **compress_kwargs)

        tdf = tdf.append(tdf0)

    return tdf


def load_google_timeline(file, user_id=None, min_accuracy_meters=None):
    """
    Load a Google Timeline trajectory in a TrajDataFrame.

    :param file: path to the file "Location History.json" inside the Google Takeout archive
    :type file: str

    :param user_id: ``str`` or ``int``
        name or ID of the user

    :param float min_accuracy_meters:
        remove points with "accuracy" value higher than ``min_accuracy_meters`` meters

    :return: TrajDataFrame
    :rtype: TrajDataFrame

    """
    # Read file
    with open(file, 'r') as f:
        goog = json.load(f)
    df = pd.DataFrame.from_dict(goog['locations'])

    # filter out inaccurate points
    if min_accuracy_meters is not None:
        df = df[df['accuracy'] < min_accuracy_meters]

    if user_id is not None:
        df['uid'] = [user_id for _ in range(len(df))]
    df['latitudeE7'] = df['latitudeE7'] / 1e7
    df['longitudeE7'] = df['longitudeE7'] / 1e7
    df['timestampMs'] = df['timestampMs'].astype(float) / 1e3

    tdf = TrajDataFrame(df, latitude='latitudeE7', longitude='longitudeE7',
                        datetime='timestampMs', timestamp=True)

    return tdf.sort_by_uid_and_datetime()



# import pandas as pd
# import numpy as np
# import osmnx as ox
# from ..utils import constants, gislib
#
# distance = gislib.getDistance
#
#
# def route(tdf, G=None, index_origin=0, index_destin=-1):
#     """Routing.
#
#     Create a route on Google Maps between two locations.
#
#     Parameters
#     ----------
#     tdf : TrajDataFrame
#         the TrajDataFrame with the locations
#
#     Examples
#     --------
#
#     G, shortest_route = route(tdf)
#
#     m = ox.plot_route_folium(G, shortest_route, route_color='green')
#
#     origin_coords = tuple(tdf1.iloc[0][['lat', 'lng']].values)
#     destin_coords = tuple(tdf1.iloc[-1][['lat', 'lng']].values)
#
#     folium.Marker(location=origin_coords,
#               icon=folium.Icon(color='red')).add_to(m)
#     folium.Marker(location=destin_coords,
#               icon=folium.Icon(color='blue')).add_to(m)
#
#     tdf1.plot_trajectory(map_f=m)
#     """
#     if index_destin == -1:
#         index_destin_last = None
#     else:
#         index_destin_last = index_destin
#     tdf1 = tdf[index_origin: index_destin_last]
#
#     origin_coords = tuple(tdf1.iloc[index_origin][[constants.LATITUDE, constants.LONGITUDE]].values)
#     destin_coords = tuple(tdf1.iloc[index_destin][[constants.LATITUDE, constants.LONGITUDE]].values)
#
#     if G is None:
#         mid_point = tuple(np.mean(np.array([origin_coords, destin_coords]), axis=0))
#         # all distances from mid_point
#         all_dists = pd.DataFrame(tdf1[[constants.LATITUDE, constants.LONGITUDE]]).apply(
#             lambda x: distance(mid_point, tuple(x.values)).m, axis=1).values
#         max_dist = 1.1 * max(all_dists)
#         G = ox.graph_from_point(mid_point, distance=max_dist)
#
#     # nodes, _ = ox.graph_to_gdfs(G)
#
#     # closest points to origin and destination on graph
#     closest_o_i = ox.utils.get_nearest_node(G, origin_coords)
#     closest_d_i = ox.utils.get_nearest_node(G, destin_coords)
#
#     # find shortest path
#     shortest_route = ox.nx.shortest_path(G, closest_o_i, closest_d_i, weight='length')
#
#     return G, shortest_route



from ..utils import gislib, utils, constants
from ..core.trajectorydataframe import *
import numpy as np
import inspect

def compress(tdf, spatial_radius_km=0.2):
    """Trajectory compression.
    
    Reduce the number of points in a trajectory for each individual in a TrajDataFrame. All points within a radius of `spatial_radius_km` kilometers from a given initial point are compressed into a single point that has the median coordinates of all points and the time of the initial point [Z2015]_.
    
    Parameters
    ----------
    tdf : TrajDataFrame
        the input trajectories of the individuals.

    spatial_radius_km : float, optional
        the minimum distance (in km) between points of the compressed trajectory. The default is `0.2`.
    
    Returns
    -------
    TrajDataFrame
        the compressed TrajDataFrame.
    
    Examples
    --------
    >>> import skmob
    >>> import pandas as pd
    >>> from skmob.preprocessing import compression
    >>> # read the trajectory data (GeoLife)
    >>> url = 'https://raw.githubusercontent.com/scikit-mobility/scikit-mobility/master/tutorial/data/geolife_sample.txt.gz'
    >>> df = pd.read_csv(url, sep=',', compression='gzip')
    >>> tdf = skmob.TrajDataFrame(df, latitude='lat', longitude='lon', user_id='user', datetime='datetime')
    >>> print(tdf.head())
             lat         lng            datetime  uid
    0  39.984094  116.319236 2008-10-23 05:53:05    1
    1  39.984198  116.319322 2008-10-23 05:53:06    1
    2  39.984224  116.319402 2008-10-23 05:53:11    1
    3  39.984211  116.319389 2008-10-23 05:53:16    1
    4  39.984217  116.319422 2008-10-23 05:53:21    1
    >>> # compress the trajectory using a spatial radius of 0.2 km
    >>> ctdf = compression.compress(tdf, spatial_radius_km=0.2)
    >>> print('Points of the original trajectory:\\t%s'%len(tdf))
    >>> print('Points of the compressed trajectory:\\t%s'%len(ctdf))
    Points of the original trajectory:	217653
    Points of the compressed trajectory:	6281
    
    References
    ----------
    .. [Z2015] Zheng, Y. (2015) Trajectory data mining: an overview. ACM Transactions on Intelligent Systems and Technology 6(3), https://dl.acm.org/citation.cfm?id=2743025
    """
    # Sort
    tdf = tdf.sort_by_uid_and_datetime()

    # Save function arguments and values in a dictionary
    frame = inspect.currentframe()
    args, _, _, arg_values = inspect.getargvalues(frame)
    arguments = dict([('function', compress.__name__)]+[(i, arg_values[i]) for i in args[1:]])

    groupby = []

    if utils.is_multi_user(tdf):
        groupby.append(constants.UID)
    if utils.is_multi_trajectory(tdf):
        groupby.append(constants.TID)

    if len(groupby) > 0:
        # Apply simplify trajectory to each group of points
        ctdf = tdf.groupby(groupby, group_keys=False).apply(_compress_trajectory, spatial_radius=spatial_radius_km)
    else:
        ctdf = _compress_trajectory(tdf, spatial_radius=spatial_radius_km)

    # TODO: remove the following line when issue #71 (Preserve the TrajDataFrame index during preprocessing operations) is solved.
    ctdf.reset_index(inplace=True, drop=True)

    ctdf.parameters = tdf.parameters
    ctdf.set_parameter(constants.COMPRESSION_PARAMS, arguments)
    return ctdf


def _compress_trajectory(tdf, spatial_radius):
    # From dataframe convert to numpy matrix
    lat_lng_dtime_other = utils.to_matrix(tdf)
    columns_order = list(tdf.columns)

    compressed_traj = _compress_array(lat_lng_dtime_other, spatial_radius)

    compressed_traj = nparray_to_trajdataframe(compressed_traj, utils.get_columns(tdf), {})
    # Put back to the original order
    compressed_traj = compressed_traj[columns_order]

    return compressed_traj


def _compress_array(lat_lng_dtime_other, spatial_radius):
    # Define the distance function to use
    measure_distance = gislib.getDistance

    compressed_traj = []
    lat_0, lon_0 = lat_lng_dtime_other[0][:2]

    sum_lat, sum_lon = [lat_0], [lon_0]
    t_0 = lat_lng_dtime_other[0][2]
    count = 1
    lendata = len(lat_lng_dtime_other) - 1

    for i in range(lendata):
        lat,lon,t = lat_lng_dtime_other[i+1][:3]

        Dr = measure_distance([lat_0,lon_0],[lat, lon])

        if Dr > spatial_radius:

            extra_cols = list(lat_lng_dtime_other[i][3:])
            compressed_traj += [[np.median(sum_lat), np.median(sum_lon), t_0] + extra_cols]

            t_0 = t
            count = 0
            lat_0, lon_0 = lat, lon
            sum_lat, sum_lon = [], []

        count += 1
        sum_lat += [lat]
        sum_lon += [lon]

    return compressed_traj
